import os
import datetime
from pathlib import Path

from ._const import (
    IFO,
    CAL_MEASUREMENT_ROOT,
    DEFAULT_CONFIG_PATH,
    DEFAULT_MODEL_PATH,
)
from . import _util
from ._util import CMDError
from ._util import logger
from ..darm import DARMModel
from ..measurement import (
    Measurement,
    ProcessSensingMeasurement,
    ProcessActuationMeasurement,
)


def list_all_measurement_sets():
    """return chronologically sorted list of all measurement ids"""
    return sorted(os.listdir(CAL_MEASUREMENT_ROOT))


def template_str_replace(template_file, replacement):
    """replace '_template_' in template file with replacement str

    """
    return os.path.basename(template_file).replace('_template_', replacement)


def find_files_from_template(directory, template):
    """find all files matching template in the specified directory

    The string '_template_' in the template name is replaced with '*',
    and file globbing is used to find files.

    """
    return sorted(list(Path(directory).glob(template_str_replace(template, '*'))))


def measurement_extract_datetime(mfile):
    """extract the datetime from a measurement file

    returns a datetime object

    """
    # FIXME: this is a BAD heuristic based on an assumed location
    # and format of the date in the file name, this should be
    # improved.
    #
    # assumes date is ISO string after final '_' separator
    ds = os.path.splitext(os.path.basename(mfile))[0].split('_')[-1]
    return _util.parse_timestamp(ds)


def find_closest_pcal(pcal_measurements, mfile):
    """find the closest pcal measurement to the given measurement file

    For each measurement file, the datetime of the measurement is
    extracted from the filename.  The closest pcal measurement and
    it's correspond datetime object are returned.

    """
    mtime = measurement_extract_datetime(mfile)
    last = (None, abs(mtime - datetime.datetime(1, 1, 1)))
    for f in pcal_measurements:
        dt = measurement_extract_datetime(f)
        diff = abs(dt - mtime)
        if diff < last[1]:
            last = (f, diff)
    return last


class MeasurementSet:
    """PyDARM Measurement Set class

    Holds information about a specific calibration measurement set.

    """

    def _find_meas_pairs(self, template, pcal_files):
        """find loop/pcal measurement pairs for a measurement matching given template

        Returns loop/pcal Measurement objects.

        """
        m_files = find_files_from_template(self.path, template)
        if len(m_files) > 1:
            logger.warning(f"found more than one measurement file for template {template}, loading latest...") # noqa E501
            m_file = m_files[-1]
        elif len(m_files) < 1:
            raise CMDError(f"no measurement found for template {template}.")
        else:
            m_file = m_files[0]
        pcal_file, pcal_dt = find_closest_pcal(pcal_files, m_file)
        # FIXME: check that pcal dt is within bounds
        return {
            'loop': Measurement(m_file),
            'pcal': Measurement(pcal_file),
        }

    def __init__(self, mset_id=None, config=DEFAULT_CONFIG_PATH):
        """load measurement set

        If `mset_id` is None or 'last'/'latest', the most recent
        measurement set will be retrieved.

        """
        self.config = _util.load_config(config)
        if mset_id in [None, 'last', 'latest']:
            logger.debug("finding last measurement set...")
            mset_list = list_all_measurement_sets()
            self.id = mset_list[-1]
        else:
            self.id = mset_id
        self.path = os.path.join(CAL_MEASUREMENT_ROOT, self.id)
        if not os.path.exists(self.path):
            raise CMDError(f"measurement set '{mset_id}' could not be found.")
        logger.info(f"measurement set: {self.path}")

        logger.debug("finding PCal measurements...")
        self.pcal_files = find_files_from_template(
            self.path,
            self.config['PCal']['template'],
        )

        self.measurements = {}

        name = 'Sensing'
        logger.debug(f"finding {name} measurements...")
        self.measurements[name] = self._find_meas_pairs(
            self.config['Sensing']['template'],
            self.pcal_files,
        )

        for stage in ['L1', 'L2', 'L3']:
            name = f'Actuation/{stage}'
            logger.debug(f"finding {name} measurements...")
            self.measurements[name] = self._find_meas_pairs(
                self.config['Actuation'][stage]['template'],
                self.pcal_files,
            )

        self.model = None
        self.processed_measurements = {}

    def load_model(self, model_ini=DEFAULT_MODEL_PATH):
        """load model and process measurements

        """
        self.model_ini = model_ini
        self.model = DARMModel(self.model_ini)
        ref_pcal = self.model.pcal.pcal_filter_bank.rstrip("_PD")

        ##########
        # load sensing

        name = 'Sensing'
        logger.info(f"processing {name}...")

        self.processed_measurements[name] = \
            ProcessSensingMeasurement(
                self.model_ini,
                self.measurements[name]['loop'],
                self.measurements[name]['pcal'],
                # FIXME: move these into the config
                (f'{IFO}:LSC-DARM1_IN2',
                 f'{IFO}:LSC-DARM1_EXC'),
                (f'{IFO}:CAL-PCALY_RX_PD_OUT_DQ',
                 f'{IFO}:LSC-DARM_IN1_DQ'),
                self.config['Sensing']['coh_thresh_loop'],
                self.config['Sensing']['coh_thresh_pcal']
            )

        ##########
        # load actuation

        self.processed_measurements['actuation'] = {}

        for stage in ['L1', 'L2', 'L3']:
            name = f'Actuation/{stage}'
            logger.info(f"processing {name}...")

            optic = self.config['Actuation'][stage]['optic']
            arm = optic[-1].lower()

            self.processed_measurements[name] = \
                ProcessActuationMeasurement(
                    self.model_ini,
                    f'actuation_{arm}_arm',
                    self.measurements[name]['loop'],
                    self.measurements[name]['pcal'],
                    # FIXME: move these into config
                    (f"{IFO}:SUS-{optic}_{stage}_CAL_EXC",
                     f'{IFO}:LSC-DARM_IN1_DQ'),
                    (f'{IFO}:CAL-{ref_pcal}_PD_OUT_DQ',
                     f'{IFO}:LSC-DARM_IN1_DQ'),
                    self.config['Actuation'][stage]['coh_thresh_loop'],
                    self.config['Actuation'][stage]['coh_thresh_pcal'],
                )

    @property
    def datetime(self):
        """return measurement set date/time as datetime object"""
        return datetime.datetime.fromisoformat(self.id)



import os

import h5py
from matplotlib import pyplot as plt

from . import _util
from ._util import logger
from ..darm import DARMModel
from ..plot import BodePlot


DATA_SAVE_FORMATS = ['.hdf5', '.h5']


def add_args(parser):
    _util.add_model_option(parser)
    _util.add_freqspec_option(parser)
    parser.add_argument(
        '--save', '-s', metavar='PATH', action='append',
        help=("save plot (.png/.pdf/.svg) or response data (.hdf5/.h5)"
              " (may be specified multiple times)"))


def main(args):
    """plot model frequency response

    """
    logger.info(f"model file: {args.model}")

    freq = _util.freq_from_spec(args.freq)

    out_data_files = set()
    out_plot_files = set()
    if args.save:
        out_files = set(args.save)
        for path in out_files:
            if os.path.splitext(path)[1] in DATA_SAVE_FORMATS:
                out_data_files.add(path)
        out_plot_files = out_files - out_data_files

    ##########

    # if args.model and os.path.splitext(args.model)[1] in DATA_SAVE_FORMATS:
    #     with h5py.File(args.model) as f:
    #         title = f.attrs.get('title', '')
    #         freq = f['Frequency'][:]
    #         G = f['Data']['DARM'][:]
    #         R = f['Data']['Response'][:]
    #         C = f['Data']['Sensing'][:]
    #         A = f['Data']['Actuation'][:]
    #         D = f['Data']['Digital'][:]

    # else:
    if True:
        DARM = DARMModel(args.model)
        title = f'{DARM.name} PyDARM model'
        G = DARM.compute_darm_olg(freq)
        R = DARM.compute_response_function(freq)
        C = DARM.sensing.compute_sensing(freq)
        A = DARM.actuation.compute_actuation(freq)
        D = DARM.digital.compute_response(freq)

    logger.info("generating bode plots...")
    bp = BodePlot(title=title)
    bp.plot(
        freq, G,
        label='DARM',
    )
    bp.plot(
        freq, C,
        label='Sensing',
        linestyle='--',
    )
    bp.plot(
        freq, A,
        label='Actuation',
        linestyle='--',
    )
    bp.plot(
        freq, D,
        label='Digital',
        linestyle='--',
    )
    bp.plot(
        freq, R,
        label='Response',
    )
    bp.legend()

    if args.save:
        for path in out_plot_files:
            logger.info(f'saving plot: {path}')
            bp.save(path)
        for path in out_data_files:
            logger.info(f'saving frequency response data: {path}')
            with h5py.File(path, 'w') as f:
                f.attrs['title'] = title
                f.create_dataset('Frequency', data=freq)
                g = f.create_group('Data')
                g.create_dataset('DARM', data=G)
                g.create_dataset('Sensing', data=C)
                g.create_dataset('Actuation', data=A)
                g.create_dataset('Digital', data=D)
                g.create_dataset('Response', data=R)
    else:
        plt.show()



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import os
import math
import json
import multiprocessing as mp
from datetime import datetime

import h5py
import emcee
import numpy as np
from scipy.signal import freqresp, ZerosPolesGain
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel

import dttxml

from .sensing import SensingModel
from .actuation import ActuationModel
from .pcal import PcalModel
from .utils import save_chain_to_hdf5, save_gpr_to_hdf5


class Measurement(object):
    """
    This class is used to get measurement data from
    a dtt measurement. This class is used in conjuction with the
    ProcessSensingMeasurement and ProcessActuationMeasurement
    class to get the input for the MCMC and GPR analysis.

    This class uses the dttxml module to import transfer functions and ASDs.

    Parameters
    ----------
    meas_file: string
        String of the name of the xml/hdf5 file of the measurement.

    Returns
    -------

    Methods (Quick reference)
    -------
    get_raw_tf: Input two channels, get array of freq and TFs
    get_raw_asd: Input one channel, get array of freq and ASD
    get_set_of_channels: No input, get a set of the channels in meas_file.

    NOTES
    -----
    If the file is not given, the class will complain.
    """

    def __init__(self, meas_file):
        if os.path.exists(meas_file) is not True:
            raise IOError('The file given does not exist.')

        self.filename = meas_file

        # Data accessing object. It is called later.
        base, ext = os.path.splitext(meas_file)
        if ext == '.xml':
            self.data_access = dttxml.DiagAccess(meas_file)
        elif ext == '.hdf5':
            self.data_access = h5py.File(meas_file, 'r')
        else:
            raise TypeError('Invalid file type input.')

        # FIXME, xml only.
        # For now, I can only access the averages and gps time from the
        # file only after I get the transfer function. There must be
        # a way to get them just after we access the file. For now,
        # I just initialize them here.
        self.averages = None
        self.gps_time = None
        self.bw = None

    def get_raw_tf(self, channelA, channelB, cohThresh=0):
        """
        Given two channels it returns the transfer function from the input
        file. Note that it gives channelB/channelA.
        In addition it populates averages and gps_time, calculates the
        uncertainty from coherence, and rejects data with coherence lower
        than the given threshold.

        Parameters
        ----------
        channelA: string
            Channel in the denominator of the TF
        channelB: string
            Channel in the numerator of the TF
        cohThresh: float
            Coherence threshold below which data from dtt are rejected

        Returns
        -------
        Returns a 4 arrays N length arrays, where N is the number of data
        points:
        Array 1: frequencies (frequencies of the data in Hz)
        Array 2: tf (transfer function of B/A)
        Array 3: coh (coherence of B/A measurement)
        Array 4: unc (the uncertainty from coherence)
        """
        set_of_A_channels, set_of_B_channels = self.get_set_of_channels()

        if (channelA in set_of_A_channels) and\
           (channelB in set_of_A_channels or channelB in set_of_B_channels):
            # Transfer function data holder. xfer is a method that comes from
            # dttxml. xfer takes channelB first and channelA second, which
            # is the opposite order that I am using. Whatever the ordering,
            # channelA always means the denominator and channelB is the
            # numerator. This should work for both Swept Sine
            # or FFT (aka broadband) measurements.
            if hasattr(self.data_access, 'xfer'):
                tf_holder = self.data_access.xfer(channelB, channelA)
                self.averages = tf_holder.averages
                self.gps_time = tf_holder.gps_second
                self.bw = tf_holder.BW
                # Frequency of the measurement in Hz
                frequencies = np.array(tf_holder.FHz)
                # Transfer function of B/A, complex array
                tf = np.array(tf_holder.xfer)
                # Coherence function of B/A, real array
                coh = np.array(tf_holder.coh)
            else:
                tf_holder = self.data_access
                self.averages = np.int32(tf_holder['parameters']['averages'][()])
                self.gps_time = np.int32(min(tf_holder[channelB][channelA]['StartTimes'][()]))
                self.bw = np.float64(tf_holder['parameters']['BW'][()])
                # Frequencies stored as strings, and need to be cast to float.
                frequencies = tf_holder[channelB][channelA]['Frequencies'][()].astype(np.float64)
                tf = tf_holder[channelB][channelA]['TransferFunctions'][()]
                coh = tf_holder[channelB][channelA]['Coherences'][()]
            # Create a map that sorts by frequency, as it is not in order by design.
            indexOrder = np.argsort(frequencies)
            frequencies = frequencies[indexOrder]
            tf = tf[indexOrder]
            coh = coh[indexOrder]
            # Uncertainty from coherence
            unc = np.sqrt((1.0-coh)/(2.0*(coh+1e-6)*self.averages))
            # Find good coherence and intersecting points
            frequencies = frequencies[(coh > cohThresh)]
            tf = tf[(coh > cohThresh)]
            unc = unc[(coh > cohThresh)]
            coh = coh[(coh > cohThresh)]
            return frequencies, tf, coh, unc
        else:
            raise ValueError('Invalid channel for transfer function.')

    def get_raw_asd(self, channelA):
        """
        It returns the ASD of channelA from the dtt file.
        This will only work for a FFT (aka broadband) type measurement.

        Parameters
        ----------
        channelA: string
            ASD channel.

        Returns
        -------
        Returns two float array:
        Array 1: frequencies (frequencies of the data in Hz)
        Array 2: asd (ASD of channelA)
        """
        set_of_A_channels, set_of_B_channels = self.get_set_of_channels()
        if (channelA in set_of_A_channels):
            # This will only work for a FFT (aka broadband) type measurement
            asd_holder = self.data_access.asd(channelA)
            self.averages = asd_holder.averages
            self.gps_time = asd_holder.gps_second
            self.bw = asd_holder.BW
            # Frequency of the measurement in Hz
            frequencies = np.array(asd_holder.FHz)
            # Amplitude Spectral Density of requested channel
            asd = np.array(asd_holder.asd)
            return frequencies, asd
        else:
            raise ValueError('Invalid channel for asd.')

    def get_set_of_channels(self):
        """
        Method to get the channels in the measurement file.
        Annoyingly, the channels are given in a set.

        For the .hdf5, channels can be swapped around freely, returing the inverse as
        expected.
        Parameters
        ----------

        Returns
        -------
        channels_A: A python set with all the A channels in the measurement
        file of this object.
        channels_B: A python set with all the B (NON-A) channels in the
        measurement file of this object.
        """
        if hasattr(self.data_access, 'channels'):
            channels_A, channels_B = self.data_access.channels()
        else:
            channels_A, channels_B = \
                    np.array(self.data_access['parameters']['channelA'][()]).astype(str), \
                    np.array(self.data_access['parameters']['channelB'][()]).astype(str)
        return channels_A, channels_B


class ProcessMeasurement(object):
    """
    This class is used in conjuction with measurement data and ouputs
    arrays used as input to MCMC and GPR routines.
    For reference, please see G1501518-v21 subway map, and specifically
    'MCMC input for C' or 'MCMC input for A'.

    Parameters
    ----------
    config_file : `str`
        This is the *.ini model file to be used to process the measurement data
    meas_sensing_or_actuation : `Measurement`
        This is the object that contains the data for either the closed
        loop gain (sensing measurement) or suspension measurement (actuation)
    meas_pcal_to_darm : `Measurement`
        This is the object that contains data for the PCAL to DARM TF
    meas_tup_1 : `str`, tuple
        This should be ('channel A', 'channel B') for the first measurement
    meas_tup_2 : `str`, tuple
        This should be ('channel A', 'channel B') for the second measurement
    meas1_cohThresh : `float`, optional
        Coherence threshold that values in measurement 1 must be above to be
        included; valid values in the range [0,1]. (default = 0)
    meas2_cohThresh : `float`, optional
        Coherence threshold that values in measurement 2 must be above to be
        included; valid values in the range [0,1]. (default = 0)
    json_results_file : `str`, optional
        Filename for a JSON file

    """

    def __init__(self, config_file, meas_sensing_or_actuation, meas_pcal_to_darm,
                 meas_tup_1, meas_tup_2, meas1_cohThresh=0, meas2_cohThresh=0,
                 json_results_file=None):
        if len(meas_tup_1) != 2:
            raise ValueError('Invalid string tuple for meas_tup_1')
        if len(meas_tup_2) != 2:
            raise ValueError('Invalid string tuple for meas_tup_2')
        if meas1_cohThresh < 0.0 or meas1_cohThresh > 1.0:
            raise ValueError('meas1_cohThresh must be in range [0,1]')
        if meas2_cohThresh < 0.0 or meas2_cohThresh > 1.0:
            raise ValueError('meas2_cohThresh must be in range [0,1]')

        self.PCAL = PcalModel(config_file)
        self.config_file = os.path.abspath(config_file)
        self.measurement_1 = meas_sensing_or_actuation
        self.measurement_2 = meas_pcal_to_darm
        self.meas_tup_1 = meas_tup_1
        self.meas_tup_2 = meas_tup_2
        self.meas1_cohThresh = meas1_cohThresh
        self.meas2_cohThresh = meas2_cohThresh
        self.json_results_file = json_results_file

        self.endstation = False
        if 'CAL-DELTAL_REF_PCAL' not in self.meas_tup_2[0]:
            self.endstation = True

    def pre_process_transfer_functions(self):
        """
        This method extracts the transfer function data using the channel names
        from the DTT files specified in the object and computes the
        intermediate data products by removing known parameters
        (see G1501518-v21).

        Returns
        -------
        frequencies : `float`, array-like
            frequency values of the transfer function, in units of Hz
        tf_meas1 : `complex128`, array-like
            the extracted transfer function from meas_tup_1, excluding
            frequency points below the meas1_cohThresh coherence threshold
        tf_meas2 : `complex128`, array-like
            the extracted transfer function from meas_tup_2, excluding
            frequency points below the meas2_cohThresh coherence threshold
        pcal_correction : `complex128`, array-like
            the transfer function for the offline photon calibrator correction
        processed_measurement_response_unc : `float`, array-like
            relative uncertainty of combined measurement data

        """

        freq_meas1, tf_meas1, coh_meas1, unc_meas1 = \
            self.measurement_1.get_raw_tf(self.meas_tup_1[0],
                                          self.meas_tup_1[1],
                                          self.meas1_cohThresh)
        freq_meas2, tf_meas2, coh_meas2, unc_meas2 = \
            self.measurement_2.get_raw_tf(self.meas_tup_2[0],
                                          self.meas_tup_2[1],
                                          self.meas2_cohThresh)

        # Merge common frequency elements from the two transfer functions.
        # Here we could have used the numpy.isin function, but there are no
        # tolerance options. It appears there may be some desire for a
        # tolerance (from other measurements), but if this is not a desireable
        # features, it could be reverted
        mask_1 = _isin(freq_meas1, freq_meas2, abs_tol=1e-3)
        mask_2 = _isin(freq_meas2, freq_meas1, abs_tol=1e-3)

        freq_meas1 = freq_meas1[mask_1]
        tf_meas1 = tf_meas1[mask_1]
        coh_meas1 = coh_meas1[mask_1]
        unc_meas1 = unc_meas1[mask_1]

        freq_meas2 = freq_meas2[mask_2]
        tf_meas2 = tf_meas2[mask_2]
        coh_meas2 = coh_meas2[mask_2]
        unc_meas2 = unc_meas2[mask_2]

        if np.allclose(freq_meas1, freq_meas2):
            frequencies = freq_meas2
        else:
            raise ValueError('Something went wrong with matching the ',
                             'frequencies of the two transfer functions')

        if len(frequencies) < 1:
            raise ValueError('There are no common frequency points for the ',
                             'two transfer functions. Maybe check coherence',
                             ' thresholds.')

        pcal_correction = self.PCAL.compute_pcal_correction(frequencies,
                                                            self.endstation)

        processed_measurement_response_unc = np.sqrt(unc_meas1**2
                                                     + unc_meas2**2)

        return frequencies, \
            tf_meas1, tf_meas2, pcal_correction, \
            processed_measurement_response_unc

    def crop_data(self, xdata, ydata, yerr, fmin=None, fmax=None):
        """
        Crop the data with optional minimum and maximum bounds

        Parameters
        ----------
        xdata : `float`, array-like
            Original frequency data
        ydata : `complex`, array-like
            Original transfer function data
        yerr : `float`, array-like
            Original uncertainties computed from the coherence values
        fmin : `float`, optional
            Optional minimum frequency value for the MCMC fitting. Default is
            the minimum frequency of xdata (default = None)
        fmax : `float`, optional
            Optional maximum frequency value for the MCMC fitting. Default is
            the maximum frequency of xdata (default = None)

        Returns
        -------
        xdata : `float`, array-like
            cropped frequencies
        ydata : `complex`, array-like
            cropped transfer function data
        yerr : `float`, array-like
            cropped uncertainties computed from the coherence values

        """
        # Mask the data the is within the frequency band selected by the user
        if fmin is not None and fmax is not None:
            xdata_mask = np.ma.masked_inside(xdata, fmin, fmax)
        elif fmin is not None and fmax is None:
            xdata_mask = np.ma.masked_greater_equal(xdata, fmin)
        elif fmin is None and fmax is not None:
            xdata_mask = np.ma.masked_less_equal(xdata, fmax)
        else:
            xdata_mask = np.ma.masked_array(xdata)
            xdata_mask.mask = True
        xdata = xdata[xdata_mask.mask]
        ydata = ydata[xdata_mask.mask]
        yerr = yerr[xdata_mask.mask]

        return xdata, ydata, yerr

    def _mcmc(self, xdata, ydata, yerr, priors, priors_bound,
              burn_in_steps=1000, steps=9000):
        """
        Run an MCMC over the processed measurement data

        Parameters
        ----------
        xdata : `float`, array-like
            Frequency values array
        ydata : `complex`, array-like
            Transfer function data corresponding to the frequency array
        yerr : `float`, array-like
            Absolute error values for each transfer function measurement point
        priors : `float`, array-like
            An N x M array of initial values where N is the
            number of walkers and M is the number of parameters
        priors_bound : `float`, array-like
            User-specified boundary of prior distribution, a 2 x M
            array where the first (second) element of each row
            corresponds to the lower (upper) bound of each parameter
        burn_in_steps : `int`, optional
            Number of steps in the burn-in phase of the MCMC (default = 1000)
        steps : `int`, optional
            Number of steps in the main phase of the MCMC (default = 9000)

        Returns
        -------
        chain : `float`, array-like
            The results of the MCMC chain

        """

        # Set the numpy parallelization to a single thread because
        # emcee has said this can cause problems
        os.environ['OMP_NUM_THREADS'] = '1'

        global xdata_gl, ydata_gl, yerr_gl
        xdata_gl, ydata_gl, yerr_gl = xdata, ydata, yerr

        # Use multiprocessing to run the MCMC faster
        # Using this instead of "with Pool() as pool:"
        # See pyDARM issue #47 for the reason for this change
        with mp.get_context('fork').Pool() as pool:
            sampler = emcee.EnsembleSampler(len(priors[:, 0]),
                                            len(priors[0, :]),
                                            _ln_prob,
                                            pool=pool,
                                            kwargs={'bound': priors_bound})
            # This is burn-in, but I wonder if there is a better way to do
            # this using the v3 of emcee. There are some optional arguments
            # when calling like 'flat','thin','discard'
            pos, prob, state = sampler.run_mcmc(priors, burn_in_steps,
                                                progress=True)
            sampler.reset()

            # Now running the actual sampling
            sampler.run_mcmc(pos, steps, progress=True)

        # I've been told by some MCMC experts that one should thin the
        # chain, but the autocorr time seems quite long, meaning we'd thin
        # the chain quite a lot unless we run a much longer MCMC chain
        # So right now we don't do any thinning
        print(sampler.get_autocorr_time(tol=10, quiet=True))

        # Return back to normal numpy operation by unsetting the environment
        os.unsetenv('OMP_NUM_THREADS')

        # Return the array of MCMC samples
        return sampler.get_chain(flat=True)

    def save_results_to_json(self, filename, fmin, fmax, mcmc_map_vals, measurement, append=False):
        """
        Save the model, input DTT filenames, coherence threshold values,
        minimum and maximum frequencies used in the MCMC, and the resulting
        MCMC MAP values

        Parameters
        ----------
        filename : str
            path and file name to JSON file
        fmin : float
            minimum frequency used for the MCMC
        fmax : float
            maximum frequency used for the MCMC
        mcmc_map_vals : `float`, array-like
            array of maximum a postiori values found from the MCMC
        measurement : str
            measurement type; one of 'sensing', 'actuation_x_arm', or 'actuation_y_arm'
        append : bool
            if True, append results to existing JSON file, otherwise overwrite

        """
        # Prepare the new dict for the JSON archive
        now = datetime.now()

        if measurement == 'sensing':
            config_dict = self.sensing.config_to_dict()
        elif (measurement == 'actuation_x_arm' or measurement == 'actuation_y_arm'):
            config_dict = self.actuation.config_to_dict()
        else:
            raise ValueError('Neither actuation nor sensing was provided')

        measurement_info = {'Measurement analysis {}'.format(
            now.strftime('%d/%m/%Y %H:%M:%S')): {
            'model_config_file': str(self.config_file),
            'meas_type': measurement,
            'model_config': config_dict,
            'loop_meas': str(self.measurement_1.filename),
            'pcal_meas': str(self.measurement_2.filename),
            'loop_tup': self.meas_tup_1,
            'pcal_tup': self.meas_tup_2,
            'loop_cohthresh': self.meas1_cohThresh,
            'pcal_cohthresh': self.meas2_cohThresh,
            'fmin': fmin, 'fmax': fmax,
            'mcmc_map_vals': mcmc_map_vals.tolist()}}

        # Either read in and make an addition to the JSON archive or create
        # the file if it didn't exist
        filename = os.path.normpath(filename)
        if os.path.exists(filename) and append is True:
            # JSON doesn't like to read/write files in append mode for some
            # reason
            with open(filename, 'r') as f:
                data = json.load(f)
            data.update(measurement_info)
            with open(filename, 'w') as f:
                json.dump(data, f)
        else:
            with open(filename, 'w') as f:
                json.dump(measurement_info, f)

        return

    def query_results_from_json(self, measurement, fmin=None, fmax=None,
                                match='latest'):
        """
        Query the JSON file for MCMC results. The query looks in the JSON
        file for a matching set of measurement files, configuration,
        coherence thresholds, and--optionally--the minimum and maximum
        frequency of the fit. `match` will determine what to return.
        It can be 'first' for the first matched, 'latest' for the last matched,
        'all' for all matched or specify which result to return by providing
        the date of measurement analysis 'DD/MM/YYYY HH:MM:SS'

        Parameters
        ----------
        measurement : str
            measurement type; one of 'sensing' or 'actuation'
        fmin : `float`, optional
            minimum frequency used for the MCMC
        fmax : `float`, optional
            maximum frequency used for the MCMC
        match : `str`, optional
            result to return; 'first', 'latest', 'all' or
            the date of measurement analysis 'DD/MM/YYYY HH:MM:SS'

        Returns
        -------
        measurement analysis date : `str`
            the date of processing measurement analysis in form of 'DD/MM/YYYY HH:MM:SS'
        mcmc_map_vals : `float`, array-like
            array of maximum a postiori values found from the MCMC
        is_pro_spring : bool
            parameter for whether the fit assumed anti-spring (False) or
            pro-spring (True)

        """

        # Load the data from the specified JSON file
        json_results_file = os.path.normpath(self.json_results_file)
        with open(json_results_file, 'r') as f:
            data = json.load(f)

        # Loop through the analyis entries, looking for only the results with
        # input values that match the request. If there are any mismatches,
        # then the loop breaks early and no result is returned
        query_match = ''
        output = []
        for idx, key in enumerate(data):
            for idx2, (key2, val2) in enumerate(data[key].items()):
                if (key2 == 'loop_meas' and
                    os.path.basename(val2) != os.path.basename(
                        self.measurement_1.filename)):
                    break
                if (key2 == 'pcal_meas' and
                    os.path.basename(val2) != os.path.basename(
                        self.measurement_2.filename)):
                    break
                if key2 == 'loop_cohthresh' and val2 != self.meas1_cohThresh:
                    break
                if key2 == 'pcal_cohthresh' and val2 != self.meas2_cohThresh:
                    break
                if fmin is not None and key2 == 'fmin' and val2 != fmin:
                    break
                if fmax is not None and key2 == 'fmax' and val2 != fmax:
                    break
                if key2 == 'model_config':
                    # copy the sensing models because we want to pop off
                    # parameters associated with coupled_cavity and
                    # detuned_spring because those will be assocaited with
                    # MCMC values
                    if measurement not in val2:
                        break

                    json_model = val2[measurement].copy()
                    if measurement == 'sensing':
                        this_model = \
                            self.sensing.config_to_dict()[measurement]
                    else:
                        this_model = \
                            self.actuation.config_to_dict()[measurement]
                    this_model_copy = this_model.copy()

                    for idx3, (key3, val3) in enumerate(
                            val2[measurement].items()):
                        if ('coupled_cavity' in key3 or
                                'detuned_spring' in key3 or
                                '_NpA' in key3 or '_NpV2' in key3):
                            json_model.pop(key3)
                    for idx3, (key3, val3) in enumerate(
                            this_model_copy.items()):
                        if ('coupled_cavity' in key3 or
                                'detuned_spring' in key3 or
                                '_NpA' in key3 or '_NpV2' in key3):
                            this_model.pop(key3)
                    if (val2['pcal'] != self.PCAL.config_to_dict()['pcal'] or
                            json_model != this_model):
                        break
                if key2 == 'mcmc_map_vals':
                    continue
                pass
            if idx2 == len(data[key])-1:
                query_match = key

                if measurement == 'sensing':
                    is_pro_spring = \
                        data[query_match]['model_config']['sensing']['is_pro_spring']
                    if is_pro_spring == 'True':
                        is_pro_spring = True
                    else:
                        is_pro_spring = False
                else:
                    is_pro_spring = None

                output.append([key[21:], np.asarray(data[query_match]['mcmc_map_vals']),
                               is_pro_spring])

            if query_match[21:] == match:
                output = output[-1]
                break
            if match == 'first' and len(output) != 0:
                break

        if match == 'latest':
            output = output[-1]

        if match != 'latest' and match != 'first' and match != 'all' and query_match[21:] != match:
            raise ValueError(
                'Need to choose what to return: first, latest, all, or measurement analysis date')

        if query_match == '':
            raise ValueError(
                'No results in the JSON archive match query values')

        return output

    def stack_measurements(self):
        """ This method is meant to be overloaded by inherited classes """

        raise Exception('This method is intended to be overloaded by child'
                        'classes')

    def _gpr(self, freq, meas_list, fmin, fmax, fmin_list, fmax_list,
             RBF_length_scale, RBF_length_scale_limits, gpr_flim):
        """
        Run a GPR over the processed measurement data

        Parameters
        ----------
        freq : `float`, array-like
            Frequencies at which to compute the GPR (Hz)
        residuals : `complex128`, array-like
            residuals of the list of measurements
        RBF_length_scale : `float`
            Initial length scale guess for the RBF term of the GPR fit (in logarithmic frequency)
        RBF_length_scale_limits : pair of floats >= 0
            The lower and upper bound on RBF_length_scale

        Returns
        -------
        y_pred : `complex128`, array-like
            Best fit curve using the GPR and covariance kernel
        sigma : `float`, array-like
            Standard deviation for the GPR
        cov : `float`, array-like
            Covariance matrix for the GPR
        stacked_meas : `list`, array-like
            list of stacked measurement frequencies, transfer functions,
            rescaled transfer functions (to reference), uncertainties, and
            residuals
        tdcfs : list
            list of time dependent correction factors for each measurement

        """

        stacked_meas, tdcfs = self.stack_measurements(meas_list, fmin, fmax,
                                                      fmin_list, fmax_list)

        for idx, this_meas in enumerate(stacked_meas):
            # Create a mask array where True means the data values should be
            # ignored while False indicates the data should be included
            if gpr_flim is not None:
                mask = np.ma.masked_outside(
                    this_meas[0], gpr_flim[0], gpr_flim[1]).mask
            else:
                mask = np.array([False] * len(this_meas[0]))

            # Create / append to an array of frequencies, residuals, and
            # uncertainties
            if idx == 0:
                residuals = np.array(
                    [this_meas[0][~mask],
                     this_meas[4][~mask],
                     this_meas[3][~mask]]).T
            else:
                residuals = np.vstack(
                    (residuals, np.array(
                        [this_meas[0][~mask],
                         this_meas[4][~mask],
                         this_meas[3][~mask]]).T))

        # Now clean up, extract, and get in the right shape for the GPR
        fdata = np.log10(np.real(residuals[None, :, 0].T))
        meas = np.hstack((np.real(residuals[None, :, 1].T), np.imag(residuals[None, :, 1].T)))
        unc = np.real(residuals[None, :, 2])

        # Gaussian process regression
        # kernel, we are still using kernel for O3
        kernel = ConstantKernel(0.5, (0.1, 2.0)) * RBF(RBF_length_scale, RBF_length_scale_limits) \
            + ConstantKernel(1.0, (0.9, 1.1))

        # Fit to data using Maximum Likelihood Estimation of the parameters
        gp = GaussianProcessRegressor(kernel=kernel, alpha=unc, n_restarts_optimizer=10)
        gp.fit(fdata, meas)

        # Make the prediction
        y_pred, sigma = gp.predict(np.log10(freq[None, :].T), return_std=True)
        y_pred, cov = gp.predict(np.log10(freq[None, :].T), return_cov=True)
        y_pred = y_pred[:, 0] + 1j*y_pred[:, 1]

        # Remove 2nd feature standard deviation and covariance contribution.
        # Since the GP fits real and imaginary as two features with the identical
        # uncertainties, the inferred covariance matrices have identical structure.
        # So we need to simply throw the second feature's identical uncertainty away.
        # This requires scikit-learn > 1.0.2.

        sigma = sigma[:, 0]
        cov = cov[:, :, 0]

        return y_pred, sigma, cov, stacked_meas, tdcfs


class ProcessSensingMeasurement(ProcessMeasurement):
    """
    This class is used in conjuction with measurement data and ouputs
    arrays used as input to MCMC and GPR routines.
    For reference, please see G1501518-v21 subway map, and specifically
    'MCMC input for C'.

    Parameters
    ----------
    config_file : `str`
        This is the *.ini model file to be used to process the measurement data
    meas_clg : `Measurement`
        This is the object that contains the data for the closed loop gain
    meas_pcal_to_darm : `Measurement`
        This is the object that contains data for the PCAL to DARM TF
    meas_tup_1 : `str`, tuple
        This should be ('channel A', 'channel B') for the first measurement
        Ex.: ('H1:LSC-DARM1_IN2', 'H1:LSC-DARM1_EXC')
    meas_tup_2 : `str`, tuple
        This should be ('channel A', 'channel B') for the second measurement
        Ex.: ('H1:CAL-PCALY_RX_PD_OUT_DQ', 'H1:LSC-DARM_IN1_DQ')
    meas1_cohThresh : `float`, optional
        Coherence threshold that values in measurement 1 must be above to be
        included; valid values in the range [0,1]. (default = 0)
    meas2_cohThresh : `float`, optional
        Coherence threshold that values in measurement 2 must be above to be
        included; valid values in the range [0,1]. (default = 0)
    json_results_file : `str`, optional
        Filename of a JSON file when stacking measurements or running GPR

    """

    def __init__(self, config_file, meas_clg, meas_pcal_to_darm, meas_tup_1,
                 meas_tup_2, meas1_cohThresh=0, meas2_cohThresh=0,
                 json_results_file=None):
        super().__init__(config_file, meas_clg, meas_pcal_to_darm,
                         meas_tup_1, meas_tup_2, meas1_cohThresh,
                         meas2_cohThresh, json_results_file)
        self.sensing = SensingModel(config_file)
        self.meas_type = 'sensing'

    def get_processed_measurement_response(self):
        """
        This method extracts the transfer function data using the channel names
        from the DTT files specified in the object and computes the measurement
        response by removing known parameters (see G1501518-v21).

        Returns
        -------
        frequencies : `float`, array-like
            frequency values of the transfer function, in units of Hz
        processed_measurement_response : `complex128`, array-like
            transfer function response of the actuation function in units
            of digital counts per meter of DARM
        processed_measurement_response_unc : `float`, array-like
            relative uncertainty of processed measurement data

        """

        # First get the intermediate data products from the measurement
        (frequencies, tf_meas1, tf_meas2, pcal_correction,
         processed_measurement_response_unc) = \
            self.pre_process_transfer_functions()

        # For the sensing function correction, we need the known terms
        tf_meas1_correction = \
            self.sensing.sensing_residual(frequencies)

        # Now bring everything together; see G1501518
        processed_measurement_response = \
            tf_meas1 * tf_meas2 / (tf_meas1_correction * pcal_correction)

        return (frequencies,
                processed_measurement_response,
                processed_measurement_response_unc)

    def rescale_sensing_by_tdcf_vals(self, frequencies, response,
                                     kappa_c=1.0, f_cc=None, **kwargs):
        """
        This method rescales the response function by dividing out the kappa_c
        and frequency response of the f_cc value and multiplying in the
        reference values given in the configuration file

        Parameters
        ----------
        frequencies : `float`, array-like
            Frequency values for the transfer functions
        response : `complex128`, array-like
            Transfer function of the optical response, having divided out the
            known quantities from the measured transfer functions
        kappa_c : `float`, optional
            The time dependent correction factor for optical gain for this
            measurement. The data is divided by this number, so if anything
            other than the default value is provided, the data will be scaled
            differently (default = 1)
        f_cc : `float`, optional
            If provided, this coupled-cavity pole frequency transfer function
            will be divided out from the data and the reference value provided
            in the configuration file will be multiplied in (default = None)

        Returns
        -------
        response : `complex128`, array-like
            The scaled transfer function of optical response

        """

        # Important note: the attribute for SensingModel are actually all
        # lowercase values. So in the ini file, if you have detuned_spring_Q
        # it is actually stored in the object as detuned_spring_q
        if f_cc is not None:
            this_response_zpk = \
                SensingModel.optical_response(
                    f_cc, self.sensing.detuned_spring_frequency,
                    self.sensing.detuned_spring_q, self.sensing.is_pro_spring)
            ref_response_zpk = \
                SensingModel.optical_response(
                    self.sensing.coupled_cavity_pole_frequency,
                    self.sensing.detuned_spring_frequency,
                    self.sensing.detuned_spring_q,
                    self.sensing.is_pro_spring)
            this_response = freqresp(this_response_zpk,
                                     2.0*np.pi*frequencies)[1]
            ref_response = freqresp(ref_response_zpk,
                                    2.0*np.pi*frequencies)[1]
            response *= ref_response / this_response

        response *= 1.0 / kappa_c

        return response

    def run_mcmc(self, fmin=None, fmax=None, priors_bound=None,
                 save_map_to_file=None, save_chain_to_file=None, **kwargs):
        """
        This method is what is used to process the sensing function measurement
        data

        Parameters
        ----------
        fmin : `float`, optional
            Optional minimum frequency value for the MCMC fitting. Default is
            the minimum frequency of xdata (default = None)
        fmax : `float`, optional
            Optional maximum frequency value for the MCMC fitting. Default is
            the maximum frequency of xdata (default = None)
        priors_bound: `float`, array-like, optional
            User-specified boundary of prior distribution. Default boundaries
            are only for spring freq (must be within this range
            0.0<spring freq<20.0) and inv Q (must be within this range
            1.e-2<inv Q<1.0).
            If choosing a different set of boundaries, then this argument
            should be in form of a 2D array, 5X2. First (second) element of
            each row corresponds to the lower (upper) bound of each
            parameters: gain, pole freq, spring freq, inv Q, residual delay.
        save_map_to_file : `str`, optional
            Filename (JSON) to save the maximum a postiori values from this
            result
        save_chain_to_file : `str`, optional
            Filename (HDF5) to save the postirior chain from this result
        kwargs : optional
            other kwargs are passed to ProcessMeasurement._mcmc

        Returns
        -------
        chain : `float`, array-like
            The MCMC chain where index:
            - 0 is the gain in units of counts per meter of DARM
            - 1 is the coupled cavity pole frequency in units of Hz
            - 2 is the optical spring frequency in units of Hz
            - 3 is the quality factor of the optical spring (Q, unitless)
            - 4 is the residual time delay in units of seconds

        """
        global meas_type_gl, is_pro_spring_gl, include_spring_gl
        meas_type_gl = self.meas_type
        is_pro_spring_gl = self.sensing.is_pro_spring

        if (hasattr(self.sensing, 'include_spring') and
                getattr(self.sensing, 'include_spring') != ''):
            # Just in case, be extra sure that we pass a boolean
            include_spring_gl = bool(self.sensing.include_spring)
        else:
            # default to True if nothing given
            include_spring_gl = True

        (frequencies, processed_measurement_response,
         processed_measurement_response_unc) = \
            self.get_processed_measurement_response()

        # For the MCMC, we need to use the absolute error and not the relative
        # uncertainty
        processed_measurement_response_err = \
            processed_measurement_response_unc * \
            np.abs(processed_measurement_response)

        # crop data used in the MCMC
        xdata, ydata, yerr = self.crop_data(frequencies,
                                            processed_measurement_response,
                                            processed_measurement_response_err,
                                            fmin, fmax)

        # priors
        nwalkers = 1000
        max_sensing_mag = np.max(np.abs(ydata))
        priors = np.hstack((
            np.random.normal([max_sensing_mag, 400.0],
                             [0.3*max_sensing_mag, 0.1*400.0],
                             size=(nwalkers, 2)),
            np.random.uniform([1., 0.01, -61.e-6], [10., 0.05, 61.e-6],
                              size=(nwalkers, 3))))

        chain = self._mcmc(xdata, ydata, yerr, priors, priors_bound, **kwargs)

        if not include_spring_gl:
            chain[:, 2] = np.zeros(len(chain[:, 0]))
            chain[:, 3] = 1e-3 * np.ones(len(chain[:, 0]))

        # Fix 1/Q ==> Q
        chain[:, 3] = 1/chain[:, 3]

        if save_map_to_file is not None:
            self.save_results_to_json(
                save_map_to_file, fmin, fmax, np.median(chain, axis=0),
                self.meas_type)

        if save_chain_to_file is not None:
            save_chain_to_hdf5(
                save_chain_to_file, self.sensing.config_to_dict(), fmin, fmax,
                self.meas_type, chain)

        return chain

    def inverse_sensing_foton_filter(self, mcmc_chain):
        """
        Prints the FOTON filter string to the terminal. Would be preferable
        at some point if this could be automatically put in to the correct
        FOTON filter file, but at present this capability is not available

        Parameters
        ----------
        mcmc_chain : `float`, array-like
            The result from the run_mcmc() method

        """

        MAP_values = np.median(mcmc_chain, axis=0)

        if not self.sensing.is_pro_spring:
            # Take s^2/(s^2 + s0^2 - i*s*s0/Q) and solve denominator for zeros
            # fs for FOTON is abs(1/(2*pi*i)*s') where s' are the two solutions
            # to the quadratic formula - see alog LHO 31693
            fs1 = (1.0/(2.0*np.pi) *
                   (-2.0*np.pi*MAP_values[2]/MAP_values[3]/2.0 +
                    0.5*np.sqrt((2.0*np.pi*MAP_values[2]/MAP_values[3])**2 +
                                4.0*(2.0*np.pi*MAP_values[2])**2)))
            fs2 = (1.0/(2.0*np.pi) *
                   (-2.0*np.pi*MAP_values[2]/MAP_values[3]/2.0 -
                    0.5*np.sqrt((2.0*np.pi*MAP_values[2]/MAP_values[3])**2 +
                                4.0*(2.0*np.pi*MAP_values[2])**2)))
        else:
            # Take s^2/(s^2 - s0^2 + i*s*s0/Q) and solve denominator for zeros
            Real = MAP_values[2]/2.0/MAP_values[3]
            Imag = (MAP_values[2]/2.0/MAP_values[3] *
                    np.sqrt(4.0*MAP_values[3]**2-1))
            fs1 = Real + Imag*1j
            fs2 = Real - Imag*1j

        # Correct normalization
        # Make a ZPK filter that is the inverse of the single pole model
        # This means the cavity pole will become a zero
        # To keep the filter constrained, we include an artificial pole at
        # 7 kHz
        singlepolefilt = ZerosPolesGain(-2.0*np.pi*MAP_values[1],
                                        -2.0*np.pi*7000.0,
                                        7000.0/MAP_values[1])
        # We want the filter normalized at 500 Hz
        normalization_value_at_500Hz = \
            abs(freqresp(singlepolefilt, 2.0*np.pi*500.0)[1][0])
        # Now compute a ZPK filter that is the inverse of the optical response
        # by including the cavity pole and (anti-)spring. To keep the filter
        # constrained, we include two poles at 0.1 Hz and one pole at 7 kHz
        modelfilt = \
            ZerosPolesGain(
                -2.0*np.pi*np.array([MAP_values[1], fs1, fs2]),
                -2.0*np.pi*np.array([0.1, 0.1, 7000]),
                np.prod(-2.0*np.pi*np.array([0.1, 0.1, 7000.0])) /
                np.prod(-2.0*np.pi*np.array([MAP_values[1], fs1, fs2])))
        # The gain is the ratio of the single pole gain to the complete optical
        # response at 500 Hz
        gainval = normalization_value_at_500Hz / \
            abs(freqresp(modelfilt, 2.0*np.pi*500.0)[1][0])

        # Print output FOTON filter
        out = 'Inverse Sensing FOTON values: '
        out += '[NB: SRCD2N zpk gain based on sensing sign in parameters '
        out += 'file]\n'
        if not self.sensing.is_pro_spring:
            out += f'SRCD2N: zpk([{MAP_values[1]:.4f};'
            out += f'{fs1:.4f};{fs2:.4f}],[0.1;0.1;7000],'
            out += f'{self.sensing.sensing_sign:.0f},"n")'
            out += f'gain({gainval:.2f})\n'
        else:
            out += f'SRCD2N: zpk([{MAP_values[1]:.4f};'
            out += f'{Real:.4f}+i*{Imag:.4f};{Real:.4f}-i*{Imag:.4f}],'
            out += f'[0.1;0.1;7000],{self.sensing.sensing_sign:.0f},"n")'
            out += f'gain({gainval:.2f})\n'
        out += f'Gain: gain({1/MAP_values[0]:.4g})\n\n'

        # Need to print a FOTON filter that has no cavity pole and no 7000 Hz
        # filter for the CFTD path
        out += 'Inverse Sensing without cavity pole FOTON values '
        out += 'for CFTD path: [NB: SRCD2N zpk gain based on sensing sign '
        out += 'in parameters file]\n'
        if not self.sensing.is_pro_spring:
            out += 'SRCD2N: '
            out += f'zpk([{fs1:.4f};{fs2:.4f}],[0.1;0.1],'
            out += f'{self.sensing.sensing_sign:.0f},"n")'
            out += f'gain({gainval:.2f})\n'
        else:
            out += f'SRCD2N: zpk([{Real:.4f}+i*{Imag:.4f};'
            out += f'{Real:.4f}-i*{Imag:.4f}],'
            out += f'[0.1;0.1],{self.sensing.sensing_sign:.0f},"n")'
            out += f'gain({gainval:.2f})\n'
        out += f'Gain: gain({1/MAP_values[0]:.4g})'

        return out

    def stack_measurements(self, measurement_list, fmin=None, fmax=None,
                           fmin_list=None, fmax_list=None):
        """
        Stack measurements together where each measurement's model for known
        filtering and digital filters is removed and the time dependence is
        removed. This makes it appear as though the measurements are taken at
        the same time. Only unknown variations or low frequency sensing spring
        variations remain.

        This object and all objects in the measurement list need to have their
        maximum a postiori values saved to the JSON file because this method
        will query the JSON file provided in each `ProcessSensingMeasurement`
        object in the list

        Parameters
        ----------
        measurement_list : `ProcessSensingMeasurement` list
            a list of measurements to normalize
        fmin : `float`, optional
            minimum frequency used for the MCMC
        fmax : `float`, optional
            maximum frequency used for the MCMC
        fmin_list : `float` list, optional
            list of minimum frequencies used for the MCMC or None
        fmax_list : `float` list, optional
            list of maximum frequencies used for the MCMC or None

        Returns
        -------
        output : `list`, array-like
            list of stacked measurement frequencies, transfer functions,
            rescaled transfer functions (to reference), uncertainties, and
            residuals
        tdcfs : list
            list of time dependent correction factors for each measurement

        """
        assert ((len(measurement_list) == len(fmin_list) and
                 len(measurement_list) == len(fmax_list)) or
                (len(measurement_list) == len(fmin_list) and
                 len(fmax_list) == 0) or
                (len(measurement_list) == len(fmax_list) and
                 len(fmin_list) == 0))

        output = []
        tdcfs = []

        ref_date, ref_map, ref_spring = self.query_results_from_json(
            self.meas_type, fmin, fmax)

        for idx, meas in enumerate(measurement_list):
            this_freq, this_resp, this_resp_unc = \
                meas.get_processed_measurement_response()
            if fmin_list is None and fmax_list is None:
                this_date, this_map, this_spring = meas.query_results_from_json(
                    self.meas_type)
            elif fmin_list is not None and fmax_list is None:
                this_date, this_map, this_spring = meas.query_results_from_json(
                    self.meas_type, fmin=fmin_list[idx])
            elif fmin_list is None and fmax_list is not None:
                this_date, this_map, this_spring = meas.query_results_from_json(
                    self.meas_type, fmax=fmax_list[idx])
            else:
                this_date, this_map, this_spring = meas.query_results_from_json(
                    self.meas_type, fmin_list[idx], fmax_list[idx])

            # We want to compute the transfer function as though we were
            # measuring the detector at the same reference time, so we need
            # to divide out the optical response from this measurement **
            # and then multiply in the optical response from the reference
            # measurement.
            # ** Note that we divide out using the optical spring f_s and Q
            # from the reference and then multiply this back in. This is
            # because we'll let the GPR take care of covering any variations
            # due to a potentially poorly modeled optical spring
            this_response_zpk = SensingModel.optical_response(
                this_map[1], ref_map[2], ref_map[3], ref_spring)
            ref_response_zpk = SensingModel.optical_response(
                ref_map[1], ref_map[2], ref_map[3], ref_spring)

            # Compute the transfer function response for the current
            # measurement model and the reference model
            this_model = this_map[0] * freqresp(
                this_response_zpk, 2.0*np.pi*this_freq)[1]
            ref_model = ref_map[0] * freqresp(
                ref_response_zpk, 2.0*np.pi*this_freq)[1]

            # The residual for this measurement is just the measured response
            # divided by the model for this measurement
            this_resid = this_resp / this_model

            # Now rescale this measurement by multiplying this measurement
            # residuals by the reference model. This is equivalent to dividing
            # out the model for this measurement and multiplying in the
            # reference model
            this_resp_r = ref_model * this_resid

            # Compute kappa_c as the ratio of the current measured gain to the
            # reference gain. We don't actually use this here other than as an
            # output
            kappa_c = this_map[0] / ref_map[0]

            output.append([this_freq, this_resp, this_resp_r, this_resp_unc, this_resid])
            tdcfs.append([kappa_c, this_map[1], ref_map[2], ref_map[3]])

        return output, tdcfs

    def run_gpr(self, frequencies, measurement_list, fmin=None, fmax=None,
                fmin_list=None, fmax_list=None,
                RBF_length_scale=1.0, RBF_length_scale_limits=(0.5, 1.5),
                gpr_flim=None, save_to_file=None):
        """
        Run a Gaussian Process Regression on a set of sensing function measurements

        Parameters
        ----------
        frequencies : `float`, array-like
            Frequencies at which to compute the GPR (Hz)
        measurement_list : `ProcessSensingMeasurement` list
            a list of measurements to normalize
        fmin : `float`, optional
            minimum frequency used for the MCMC of reference measurement
        fmax : `float`, optional
            maximum frequency used for the MCMC of reference measurement
        fmin_list : `float` list, optional
            list of minimum frequencies used for the MCMC or None
        fmax_list : `float` list, optional
            list of maximum frequencies used for the MCMC or None
        RBF_length_scale : `float`, optional
            Initial length scale guess for the RBF term of the GPR fit (in logarithmic frequency)
        RBF_length_scale_limits : pair of floats >= 0, optional
            The lower and upper bound on RBF_length_scale
        gpr_flim : `tuple`, optional
            Minimum and maximum frequency points to include in the GPR, (fmin, fmax)
        save_to_file : str, optional
            Filename (HDF5) to save the data from this result

        Returns
        -------
        y_pred : `complex128`, array-like
            Best fit curve using the GPR and covariance kernel
        sigma : `float`, array-like
            Standard deviation for the GPR
        cov : `float`, array-like
            Covariance matrix for the GPR
        stacked_meas : `list`, array-like
            list of stacked measurement frequencies, transfer functions,
            rescaled transfer functions (to reference), uncertainties, and
            residuals
        tdcfs : list
            list of time dependent correction factors for each measurement

        """
        assert ((len(measurement_list) == len(fmin_list) and
                 len(measurement_list) == len(fmax_list)) or
                (len(measurement_list) == len(fmin_list) and
                 len(fmax_list) == 0) or
                (len(measurement_list) == len(fmax_list) and
                 len(fmin_list) == 0))

        y_pred, sigma, cov, stacked_meas, tdcfs = self._gpr(
            frequencies, measurement_list, fmin, fmax,
            fmin_list, fmax_list, RBF_length_scale,
            RBF_length_scale_limits, gpr_flim)

        # save the result to HDF5
        if save_to_file is not None:
            save_gpr_to_hdf5(save_to_file, self.sensing, 'sensing', y_pred,
                             cov, frequencies)

        return y_pred, sigma, cov, stacked_meas, tdcfs


class ProcessActuationMeasurement(ProcessMeasurement):
    """
    This class is used in conjuction with measurement data and ouputs
    arrays used as input to MCMC and GPR routines.
    For reference, please see G1501518-v21 subway map, and specifically
    'MCMC input for A'.

    Parameters
    ----------
    config_file : `str`
        This is the *.ini model file to be used to process the measurement data
    meas_type : `str`
        This should be one of 'actuation_x_arm' or 'actuation_y_arm'
    meas_sus_exc_to_darm : `Measurement`
        This is the object that contains the data for the DARM_IN channel to
        suspension excitation response
    meas_pcal_to_darm : `Measurement`
        This is the object that contains data for the PCAL to DARM TF
    meas_tup_1 : `str`, tuple
        This should be ('channel A', 'channel B') for the first measurement
        Ex.: ('H1:SUS-ETMX_L3_CAL_EXC', 'H1:LSC-DARM1_IN1_DQ')
    meas_tup_2 : `str`, tuple
        This should be ('channel A', 'channel B') for the second measurement
        Ex.: ('H1:CAL-PCALY_RX_PD_OUT_DQ', 'H1:LSC-DARM_IN1_DQ')
    meas1_cohThresh : `float`, optional
        Coherence threshold that values in measurement 1 must be above to be
        included; valid values in the range [0,1]. (default = 0)
    meas2_cohThresh : `float`, optional
        Coherence threshold that values in measurement 2 must be above to be
        included; valid values in the range [0,1]. (default = 0)
    json_results_file : `str`, optional
        Filename of a JSON file when stacking measurements or running GPR

    """

    def __init__(self, config_file, meas_type, meas_sus_exc_to_darm, meas_pcal_to_darm,
                 meas_tup_1, meas_tup_2, meas1_cohThresh=0, meas2_cohThresh=0,
                 json_results_file=None):
        if not (meas_type == 'actuation_x_arm' or
                meas_type == 'actuation_y_arm'):
            raise ValueError('Invalid meas_type')
        super().__init__(config_file, meas_sus_exc_to_darm, meas_pcal_to_darm,
                         meas_tup_1, meas_tup_2, meas1_cohThresh,
                         meas2_cohThresh, json_results_file)
        self.actuation = ActuationModel(config_file, measurement=meas_type)
        self.meas_type = meas_type

    def get_processed_measurement_response(self):
        """
        This method extracts the transfer function data using the channel names
        from the DTT files specified in the object and computes the measurement
        response by removing known parameters (see G1501518-v21).

        Returns
        -------
        frequencies : `float`, array-like
            frequency values of the transfer function, in units of Hz
        processed_measurement_response : `complex128`, array-like
            transfer function response of the actuation function, in units of
            newtons per driver output (amps or volts**2). For example, for TST
            the units will be in N/V**2
        processed_measurement_response_unc : `float`, array-like
            relative uncertainty of processed measurement data

        """

        # First get the intermediate data products from the measurement
        (frequencies, tf_meas1, tf_meas2, pcal_correction,
         processed_measurement_response_unc) = \
            self.pre_process_transfer_functions()

        # For the actuation function correction, we need the known terms
        correction_uim, correction_pum, correction_tst = \
            self.actuation.known_actuation_terms_for_measurement(frequencies)

        # Now bring everything together; see G1501518
        # Choose which transfer function based on the user input
        if 'L3' in self.meas_tup_1[0].split(':')[1]:
            processed_measurement_response = \
                (tf_meas1 / tf_meas2) * (pcal_correction / correction_tst)
        elif 'L2' in self.meas_tup_1[0].split(':')[1]:
            processed_measurement_response = \
                (tf_meas1 / tf_meas2) * (pcal_correction / correction_pum)
        elif 'L1' in self.meas_tup_1[0].split(':')[1]:
            processed_measurement_response = \
                (tf_meas1 / tf_meas2) * (pcal_correction / correction_uim)
        else:
            raise ValueError('Could not find L1, L2, or L3 in the',
                             'meas_tup_1 string')

        return (frequencies,
                processed_measurement_response,
                processed_measurement_response_unc)

    def rescale_actuation_by_tdcf_val(self, response, kappa_Ai=1.0, **kwargs):
        """
        This method rescales the response function by dividing out the kappa_Ai
        value

        Parameters
        ----------
        response : `complex128`, array-like
            Transfer function of the optical response, having divided out the
            known quantities from the measured transfer functions
        kappa_Ai : `float`, optional
            The time dependent correction factor for actuator gain for this
            measurement. The data is divided by this number, so if anything
            other than the default value is provided, the data will be scaled
            differently (default = 1)

        """

        return response / kappa_Ai

    def run_mcmc(self, fmin=None, fmax=None, priors_bound=None,
                 save_map_to_file=None, save_chain_to_file=None, **kwargs):
        """
        This method is what is used to process the actuation function
        measurement data

        Parameters
        ----------
        fmin : `float`, optional
            Optional minimum frequency value for the MCMC fitting. Default is
            the minimum frequency of xdata (default = None)
        fmax : `float`, optional
            Optional maximum frequency value for the MCMC fitting. Default is
            the maximum frequency of xdata (default = None)
        priors_bound: `float`, array-like, optional
            User-specified boundary of prior distribution. Default boundary
            is only for gain, gain must be larger or equal to 0.0.
            If choosing a different set of boundaries, then this argument
            should be in form of a 2D array, 2X2. First (second) element of
            each row corresponds to the lower (upper) bound of each
            parameters: gain, residual delay.
        save_map_to_file : `str`, optional
            Filename (JSON) to save the maximum a postiori values from this
            result
        save_chain_to_file : `str`, optional
            Filename (HDF5) to save the postirior chain from this result
        kwargs : optional
            other kwargs are passed to ProcessMeasurement._mcmc

        Returns
        -------
        chain : `float`, array-like
            The MCMC chain where index:
            - 0 is the gain in units of newtons per driver output (amps or
            volts**2). For example, for TST the units will be in N/V**2.
            - 1 is the residual time delay in units of seconds

        """
        global meas_type_gl
        meas_type_gl = self.meas_type

        (frequencies, processed_measurement_response,
         processed_measurement_response_unc) = \
            self.get_processed_measurement_response()

        # For the MCMC, we need to use the absolute error and not the relative
        # uncertainty
        processed_measurement_response_err = \
            processed_measurement_response_unc * \
            np.abs(processed_measurement_response)

        # crop data used in the MCMC
        xdata, ydata, yerr = self.crop_data(frequencies,
                                            processed_measurement_response,
                                            processed_measurement_response_err,
                                            fmin, fmax)

        # priors
        nwalkers = 1000
        mean_actuation_mag = np.mean(np.abs(ydata))
        priors = np.hstack((
            np.random.normal(mean_actuation_mag, 0.3*mean_actuation_mag,
                             size=(nwalkers, 1)),
            np.random.uniform(-61.e-6, 61.e-6, size=(nwalkers, 1))))

        chain = self._mcmc(xdata, ydata, yerr, priors, priors_bound, **kwargs)

        if save_map_to_file is not None:
            self.save_results_to_json(
                save_map_to_file, fmin, fmax, np.median(chain, axis=0),
                self.meas_type)

        if save_chain_to_file is not None:
            save_chain_to_hdf5(
                save_chain_to_file, self.actuation.config_to_dict(), fmin, fmax,
                self.meas_type, chain)

        return chain

    def stack_measurements(self, measurement_list, fmin=None, fmax=None,
                           fmin_list=None, fmax_list=None, use_ini_reference=True):
        """
        Stack measurements together where each measurement's model for known
        filtering and digital filters is removed and the time dependence is
        removed. This makes it appear as though the measurements are taken at
        the same time. Only unknown variations or low frequency sensing spring
        variations remain.

        This object and all objects in the measurement list need to have their
        maximum a postiori values saved to the JSON file because this method
        will query the JSON file provided in each `ProcessSensingMeasurement`
        object in the list

        Parameters
        ----------
        measurement_list : `ProcessSensingMeasurement` list
            a list of measurements to normalize
        fmin : `float`, optional
            minimum frequency used for the MCMC
        fmax : `float`, optional
            maximum frequency used for the MCMC
        fmin_list : `float` list, optional
            list of minimum frequencies used for the MCMC or None
        fmax_list : `float` list, optional
            list of maximum frequencies used for the MCMC or None
        use_ini_reference : `boolean'
            flag to use the parameters for the reference model
            from the ini (default = True)

        Returns
        -------
        output : `list`, array-like
            list of stacked measurement frequencies, transfer functions,
            rescaled transfer functions (to reference), uncertainties, and
            residuals
        tdcfs : list
            list of time dependent correction factors for each measurement

        """
        assert ((len(measurement_list) == len(fmin_list) and
                 len(measurement_list) == len(fmax_list)) or
                (len(measurement_list) == len(fmin_list) and
                 len(fmax_list) == 0) or
                (len(measurement_list) == len(fmax_list) and
                 len(fmin_list) == 0))

        output = []
        tdcfs = []

        ref_map = self.query_results_from_json(self.meas_type, fmin, fmax)[1]

        for idx, meas in enumerate(measurement_list):
            this_freq, this_resp, this_resp_unc = \
                meas.get_processed_measurement_response()
            if fmin_list is None and fmax_list is None:
                this_map = meas.query_results_from_json(self.meas_type)[1]
            elif fmin_list is not None and fmax_list is None:
                this_map = meas.query_results_from_json(
                    self.meas_type, fmin=fmin_list[idx])[1]
            elif fmin_list is None and fmax_list is not None:
                this_map = meas.query_results_from_json(
                    self.meas_type, fmax=fmax_list[idx])[1]
            else:
                this_map = meas.query_results_from_json(
                    self.meas_type, fmin_list[idx], fmax_list[idx])[1]

            # We want to compute the transfer function as though we were
            # measuring the detector at the same reference time, so we need
            # to divide out the actuation response from this measurement **
            # and then multiply in the actuation response from the reference
            # measurement.
            # ** Note that we only scale by the gain and not any residual time
            # delay
            this_model = this_map[0]
            ref_model = ref_map[0]

            # The residual for this measurement is just the measured response
            # divided by the model for this measurement
            this_resid = this_resp / this_model

            # Now rescale this measurement by multiplying this measurement
            # residuals by the reference model. This is equivalent to dividing
            # out the model for this measurement and multiplying in the
            # reference model
            this_resp_r = ref_model * this_resid

            # Compute kappa_a as the ratio of the current measured gain to the
            # reference gain. We don't actually use this here other than as an
            # output
            kappa_a = this_map[0] / ref_map[0]

            output.append([this_freq, this_resp, this_resp_r, this_resp_unc, this_resid])
            tdcfs.append(kappa_a)

        return output, tdcfs

    def run_gpr(self, frequencies, measurement_list, fmin=None, fmax=None,
                fmin_list=None, fmax_list=None,
                RBF_length_scale=1.0, RBF_length_scale_limits=(0.5, 1.5),
                gpr_flim=None, save_to_file=None):
        """
        Run a Gaussian Process Regression on a set of actuation function measurements

        Parameters
        ----------
        frequencies : `float`, array-like
            Frequencies at which to compute the GPR (Hz)
        measurement_list : `ProcessSensingMeasurement` list
            a list of measurements to normalize
        fmin : `float`, optional
            minimum frequency used for the MCMC of reference measurement
        fmax : `float`, optional
            maximum frequency used for the MCMC of reference measurement
        fmin_list : `float` list, optional
            list of minimum frequencies used for the MCMC or None
        fmax_list : `float` list, optional
            list of maximum frequencies used for the MCMC or None
        RBF_length_scale : `float`, optional
            Initial length scale guess for the RBF term of the GPR fit (in logarithmic frequency)
        RBF_length_scale_limits : pair of floats >= 0, optional
            The lower and upper bound on RBF_length_scale
        gpr_flim : `tuple`, optional
            Minimum and maximum frequency points to include in the GPR, (fmin, fmax)
        save_to_file : str, optional
            Filename (HDF5) to save the data from this result

        Returns
        -------
        y_pred : `complex128`, array-like
            Best fit curve using the GPR and covariance kernel
        sigma : `float`, array-like
            Standard deviation for the GPR
        cov : `float`, array-like
            Covariance matrix for the GPR
        stacked_meas : `list`, array-like
            list of stacked measurement frequencies, transfer functions,
            rescaled transfer functions (to reference), uncertainties, and
            residuals
        tdcfs : list
            list of time dependent correction factors for each measurement

        """
        assert ((len(measurement_list) == len(fmin_list) and
                 len(measurement_list) == len(fmax_list)) or
                (len(measurement_list) == len(fmin_list) and
                 len(fmax_list) == 0) or
                (len(measurement_list) == len(fmax_list) and
                 len(fmin_list) == 0))

        y_pred, sigma, cov, stacked_meas, tdcfs = self._gpr(
            frequencies, measurement_list, fmin, fmax,
            fmin_list, fmax_list, RBF_length_scale,
            RBF_length_scale_limits, gpr_flim)

        # save the result to HDF5
        if save_to_file is not None:
            save_gpr_to_hdf5(save_to_file, self.actuation, self.meas_type,
                             y_pred, cov, frequencies)

        return y_pred, sigma, cov, stacked_meas, tdcfs


def _ln_prob(pars, bound=None):
    """
    This is a helper function for the MCMC analysis and not intended for
    general use

    See https://emcee.readthedocs.io

    """

    if meas_type_gl == 'sensing':
        if include_spring_gl:
            if bound is None:
                if (pars[2] < 0.0 or pars[2] > 20.0 or pars[3] < 1e-2 or
                        pars[3] > 1.0):
                    return -np.inf
            else:
                if (pars[0] < bound[0][0] or pars[0] > bound[0][1] or
                        pars[1] < bound[1][0] or pars[1] > bound[1][1] or
                        pars[2] < bound[2][0] or pars[2] > bound[2][1] or
                        pars[3] < bound[3][0] or pars[3] > bound[3][1] or
                        pars[4] < bound[4][0] or pars[4] > bound[4][1]):
                    return -np.inf

            this_response = \
                pars[0] * freqresp(SensingModel.optical_response(
                    pars[1], pars[2], 1/pars[3], is_pro_spring_gl),
                    2.0*np.pi*xdata_gl)[1] * np.exp(-2.*np.pi*1j*pars[4] * xdata_gl)

        else:
            if bound is not None:
                if (pars[0] < bound[0][0] or pars[0] > bound[0][1] or
                        pars[1] < bound[1][0] or pars[1] > bound[1][1] or
                        pars[4] < bound[4][0] or pars[4] > bound[4][1]):
                    return -np.inf

            this_response = \
                pars[0] * freqresp(SensingModel.optical_response(
                    pars[1], 0, 1e-3, is_pro_spring_gl),
                    2.0*np.pi*xdata_gl)[1] * np.exp(-2.*np.pi*1j*pars[4] * xdata_gl)

    elif (meas_type_gl == 'actuation_x_arm'
          or meas_type_gl == 'actuation_y_arm'):
        if bound is None:
            if pars[0] <= 0.0:
                return -np.inf
        else:
            if (pars[0] < bound[0][0] or pars[0] > bound[0][1] or
                    pars[1] < bound[1][0] or pars[1] > bound[1][1]):
                return -np.inf
        this_response = (pars[0] *
                         np.exp(-2.*np.pi*1j*pars[1] * xdata_gl))
    return np.sum(_log_norm_pdf(ydata_gl, this_response, yerr_gl))


def _log_norm_pdf(x, mu, sigma):
    """
    This is a helper function for the MCMC and not intended for general use

    This calculates the log of a Whittle likelihood function for two dimensions,
    PDF(x|mu,sigma) = (2*pi*sigma**2) * exp[-(x-mu)**2/(2*sigma**2)], where the
    two dimensions originate from the real and imaginary contributions from the
    data. We assume that the uncertainty in the two dimensions is uncorrelated,
    and that sigma is the standard deviation for the Gaussian in each of the two
    dimensions. Here `x` are the data values, `mu` are the expected values,
    and `sigma` are the absolute error values (not relative error)

    """
    return -0.5*abs((x-mu)/sigma)**2 - np.log(2.0*np.pi*sigma**2)


def _isin(element, test_elements, rel_tol=1e-09, abs_tol=0.0):
    """
    This function is related to the numpy function isin, but here we allow
    for a user-specified tolerance based on math.isclose()
    Calculates element in test_elements, broadcasting over element only.
    Returns a boolean array of the same shape as element that is True where
    an element of element is in test_elements and False otherwise.

    We implement and use this function because apparently there are sometimes
    small numerical errors on the frequency values extracted from DTT files
    leading to errors when combining transfer functions from different
    measurements. It is not intended for general use.

    Parameters
    ----------
    element : `float`, array-like
        Input array
    test_elements : `float`, array-like
        The values against which to test each value of element
    rel_tol : float
        relative tolerance
    abs_tol : float
        minimum absolute tolerance - useful for comparisons near zero.
        abs_tol must be at least zero

    Returns
    -------
    isin : `boolean`, array-like
        Has the same shape as element. The values element[isin] are in
        test_elements

    """

    isin = np.empty(element.shape, dtype=bool)
    for i in range(len(element)):
        isin[i] = False
        for j in range(len(test_elements)):
            if math.isclose(element[i], test_elements[j], rel_tol=rel_tol,
                            abs_tol=abs_tol):
                isin[i] = True
                break

    return isin



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

from scipy import signal, io
import numpy as np


def analog_aa_or_ai_filter_response(mat_file, frequencies):
    """
    This comes from an export of a Matlab LTI object which was a fit of
    measurements made at LHO and LLO by JeffK, JoeB, et al. The object has
    zeros, poles, gain, and a delay, so the exported mat-file has the arrays
    of zeros and poles, gain, and delay in the file. Measurements were made
    around ER7. TODO: find alog entries?

    Parameters
    ----------
    mat_file : `str`
        path and filename of the Matlab mat-file export
    frequencies : `float`, array-like
        array of frequencies to compute the response

    Returns
    -------
    tf : `complex128`, array-like
        transfer function response of the analog AA or AI filter
    """

    mat = io.loadmat(mat_file)
    analog_aa_or_ai_zpk = signal.ZerosPolesGain(mat['AAzeros'][:, 0],
                                                mat['AApoles'][:, 0],
                                                mat['AAgain'][0, 0])
    analog_aa_or_ai_delay = np.exp(
        -2.0*np.pi*mat['AAdelay'][0, 0]*1j*frequencies)

    return signal.freqresp(analog_aa_or_ai_zpk,
                           2.0*np.pi*frequencies)[1] * analog_aa_or_ai_delay



import unittest
import pydarm
import numpy as np


class TestAnalogDriverResponse(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_response = {
            'UIM': {
                'UL': np.ones(len(self.frequencies), dtype='complex128'),
                'LL': np.ones(len(self.frequencies), dtype='complex128'),
                'UR': np.ones(len(self.frequencies), dtype='complex128'),
                'LR': np.ones(len(self.frequencies), dtype='complex128')},
            'PUM': {
                'UL': np.ones(len(self.frequencies), dtype='complex128'),
                'LL': np.ones(len(self.frequencies), dtype='complex128'),
                'UR': np.ones(len(self.frequencies), dtype='complex128'),
                'LR': np.ones(len(self.frequencies), dtype='complex128')},
            'TST': {
                'UL': np.array(
                    [0.9999998948885249-0.0003352715041293239j,
                     0.9999993023409579-0.0008637599652766293j,
                     0.9999953694243224-0.002225297865133263j,
                     0.9999692660865103-0.005732898947830594j,
                     0.9997960415022515-0.014767298626171824j,
                     0.9986476944616459-0.03800441737197289j,
                     0.9910869734117239-0.09722224905346574j,
                     0.9434615871897538-0.23929884983400718j,
                     0.7100118921447782-0.47534426423585624j,
                     0.23350097886085233-0.48059520727971594j]),
                'LL': np.array(
                    [0.9999998916146684-0.00034119500566379466j,
                     0.9999992806113166-0.0008790206699527335j,
                     0.9999952251984726-0.002264613688569092j,
                     0.9999683088528678-0.0058341823535449635j,
                     0.999789689822471-0.01502813255515821j,
                     0.9986056151986867-0.03867465221930742j,
                     0.990811116442425-0.09891939009102466j,
                     0.9417711330516-0.2432045844267566j,
                     0.7028366049212413-0.48037909581270577j,
                     0.22293976812324173-0.47855091024022334j]),
                'UR': np.ones(len(self.frequencies), dtype='complex128'),
                'LR': np.ones(len(self.frequencies), dtype='complex128')}}

    def tearDown(self):
        del self.frequencies
        del self.known_response

    def test_analog_driver_response(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR =
tst_driver_uncompensated_Z_LR =
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR =
tst_driver_uncompensated_P_LR =
''',
                                            measurement='actuation_x_arm')

        test_response = A.analog_driver_response(self.frequencies)
        for i, stage in enumerate(test_response.keys()):
            for j, quadrant in enumerate(test_response[stage].keys()):
                for k, val in enumerate(test_response[stage][quadrant]):
                    self.assertAlmostEqual(
                        np.abs(val),
                        np.abs(self.known_response[stage][quadrant][k]))
                    self.assertAlmostEqual(
                        np.angle(val, deg=True),
                        np.angle(self.known_response[stage][quadrant][k], deg=True))


class TestUIMDCGainApct(unittest.TestCase):

    def setUp(self):
        # uim_driver_DC_trans_ApV = 6.1535e-4
        # dac_gain = 7.62939453125e-05
        self.known_uim_dc_gain = 4.694747924804687e-08

    def tearDown(self):
        del self.known_uim_dc_gain

    def test_uim_dc_gain_Apct(self):
        """ Test the UIM DC gain A / ct """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
uim_driver_DC_trans_ApV = 6.1535e-4
dac_gain = 7.62939453125e-05
''',
                                            measurement='actuation_x_arm')
        uim_gain = A.uim_dc_gain_Apct()
        self.assertAlmostEqual(uim_gain, self.known_uim_dc_gain)


class TestPUMDCGainApct(unittest.TestCase):

    def setUp(self):
        # pum_driver_DC_trans_ApV = 2.6847e-4
        # dac_gain = 7.62939453125e-05
        # pum_coil_outf_signflip = -1
        self.known_pum_dc_gain = -2.0482635498046874e-08

    def tearDown(self):
        del self.known_pum_dc_gain

    def test_pum_dc_gain_Apct(self):
        """ Test the PUM DC gain A / ct """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
pum_driver_DC_trans_ApV = 2.6847e-4
dac_gain = 7.62939453125e-05
pum_coil_outf_signflip = -1
''',
                                            measurement='actuation_x_arm')
        pum_gain = A.pum_dc_gain_Apct()
        self.assertAlmostEqual(pum_gain, self.known_pum_dc_gain)


class TestTSTDCGainV2pct(unittest.TestCase):

    def setUp(self):
        # tst_driver_DC_gain_VpV_HV = 40
        # tst_driver_DC_gain_VpV_LV = 1.881
        # dac_gain = 7.62939453125e-05
        # actuation_esd_bias_voltage = -9.3
        self.known_tst_dc_gain = -0.10677062988281251

    def tearDown(self):
        del self.known_tst_dc_gain

    def test_tst_dc_gain_Apct(self):
        """ Test the TST DC gain V**2 / ct """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
actuation_esd_bias_voltage = -9.3
dac_gain = 7.62939453125e-05
''',
                                            measurement='actuation_x_arm')
        tst_gain = A.tst_dc_gain_V2pct()
        self.assertAlmostEqual(tst_gain, self.known_tst_dc_gain)


class TestUIMDCGainNpct(unittest.TestCase):

    def setUp(self):
        # uim_driver_DC_trans_ApV = 6.1535e-4
        # dac_gain = 7.62939453125e-05
        # uim_NpA       = 1.634
        self.known_uim_dc_gain = 7.671218109130858e-08

    def tearDown(self):
        del self.known_uim_dc_gain

    def test_uim_dc_gain_Apct(self):
        """ Test the UIM DC gain N / ct """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
uim_driver_DC_trans_ApV = 6.1535e-4
dac_gain = 7.62939453125e-05
uim_NpA = 1.634
''',
                                            measurement='actuation_x_arm')
        uim_gain = A.uim_dc_gain_Npct()
        self.assertAlmostEqual(uim_gain, self.known_uim_dc_gain)


class TestPUMDCGainNpct(unittest.TestCase):

    def setUp(self):
        # pum_driver_DC_trans_ApV = 2.6847e-4
        # dac_gain = 7.62939453125e-05
        # pum_coil_outf_signflip = -1
        # pum_NpA       = 0.02947
        self.known_pum_dc_gain = -6.036232681274414e-10

    def tearDown(self):
        del self.known_pum_dc_gain

    def test_pum_dc_gain_Npct(self):
        """ Test the PUM DC gain N / ct """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
pum_driver_DC_trans_ApV = 2.6847e-4
dac_gain = 7.62939453125e-05
pum_NpA = 0.02947
pum_coil_outf_signflip = -1
''',
                                            measurement='actuation_x_arm')
        pum_gain = A.pum_dc_gain_Npct()
        self.assertAlmostEqual(pum_gain, self.known_pum_dc_gain)


class TestTSTDCGainNpct(unittest.TestCase):

    def setUp(self):
        # tst_driver_DC_gain_VpV_HV = 40
        # tst_driver_DC_gain_VpV_LV = 1.881
        # dac_gain = 7.62939453125e-05
        # actuation_esd_bias_voltage = -9.3
        # tst_NpV2    = 4.427e-11
        self.known_tst_dc_gain = -4.72673578491211e-12

    def tearDown(self):
        del self.known_tst_dc_gain

    def test_tst_dc_gain_Npct(self):
        """ Test the TST DC gain N / ct """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
actuation_esd_bias_voltage = -9.3
dac_gain = 7.62939453125e-05
tst_NpV2 = 4.427e-11
''',
                                            measurement='actuation_x_arm')
        tst_gain = A.tst_dc_gain_Npct()
        self.assertAlmostEqual(tst_gain, self.known_tst_dc_gain)


class TestSUSDigitalFilters(unittest.TestCase):

    def setUp(self):
        # Pre-defined values
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.tst_isc_inf_filter_response = np.array(
            [1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j])
        self.tst_lock_filter_response = np.array(
            [0.999995665359871-0.000272790003103j,
             0.999995496265676-0.000702789632297j,
             0.999994373922495-0.001810604217536j,
             0.999986924147203-0.004664770123548j,
             0.999937459270884-0.012019700386180j,
             0.999608338031661-0.030997995364512j,
             0.997387781574251-0.080406381478687j,
             0.980886105795703-0.217157945258743j,
             0.638188154184932-0.898259426552110j,
             0.707847556914040+0.798465985670971j])
        self.tst_drive_align_filter_response = np.array(
            [-0.031382125001088+0.305163415820658j,
             2.425639846913828+0.541266964013526j,
             4.772436789791023-14.431858429029758j,
             -10.799493075302230-21.659812735113267j,
             -31.182515810388480-18.548259489461532j,
             -39.526836812833459-6.894921607809322j,
             -36.129734307732122-0.266960569021572j,
             -35.682495450563579-0.320690652602195j,
             -35.696331063875391-0.134563202055048j,
             -35.699846481043664-0.037488662354112j])
        self.pum_lock_filter_response = np.array(
            [-0.157592900352767+0.127863513836283j,
             -2.826004221047697+0.243072450878951j,
             -28.982811336208435+1.154255573110084j,
             -199.257228525459169+19.053501528584700j,
             -1309.778579816084402+329.356447710530347j,
             -7333.646298110769749+6093.483160258686439j,
             29683.804497749919392+27721.559707795226132j,
             29263.012403179542162+14747.370377871528035j,
             26716.549838518367324-14947.643675827408515j,
             3964.230336695232381-11028.505558133558225j])
        self.pum_drive_align_filter_response = np.array(
            [1.007811710686028-0.002244893005650j,
             1.061673400011530-0.014847884581376j,
             1.072155888474460-0.422340584495491j,
             0.652165100671343-0.605111854260755j,
             0.083870120634454-0.507781585073636j,
             -0.137921317637276-0.162320031347624j,
             -0.011808103201744+0.002762524811938j,
             0.000584567255433-0.005925249411234j,
             0.001359953413426-0.002900713040568j,
             0.000094860790992-0.000924075886957j])
        self.uim_lock_filter_response = np.array(
            [0.003812890441709-0.010711971680203j,
             0.005023211450190+0.005667020241205j,
             0.014779885019662+0.027004020822383j,
             0.086070868465663+0.022181985878403j,
             0.022327336104072-0.040946207260487j,
             0.002880776234144-0.016580573534908j,
             0.000421760000215-0.006429875356407j,
             0.000062535339517-0.002480687522224j,
             0.000008679381355-0.000924441210796j,
             0.000000651564638-0.000253298362886j])
        self.uim_drive_align_filter_response = np.array(
            [1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j,
             1.000000000000000+0.000000000000000j])

    def tearDown(self):
        del self.frequencies
        del self.tst_isc_inf_filter_response
        del self.tst_lock_filter_response
        del self.tst_drive_align_filter_response
        del self.pum_lock_filter_response
        del self.pum_drive_align_filter_response
        del self.uim_lock_filter_response
        del self.uim_drive_align_filter_response

    def test_sus_digital_filters_response(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0
''',
                                            measurement='actuation_x_arm')
        [tst_isc_inf_filter_response, tst_lock_filter_response,
         tst_drive_align_filter_response, pum_lock_filter_response,
         pum_drive_align_filter_response, uim_lock_filter_response,
         uim_drive_align_filter_response] = \
            A.sus_digital_filters(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(tst_isc_inf_filter_response[n]),
                                   np.abs(self.tst_isc_inf_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(tst_isc_inf_filter_response[n], deg=True),
                np.angle(self.tst_isc_inf_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(tst_lock_filter_response[n]),
                                   np.abs(self.tst_lock_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(tst_lock_filter_response[n], deg=True),
                np.angle(self.tst_lock_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(tst_drive_align_filter_response[n]),
                                   np.abs(
                                       self.tst_drive_align_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(tst_drive_align_filter_response[n], deg=True),
                np.angle(self.tst_drive_align_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(pum_lock_filter_response[n]),
                                   np.abs(self.pum_lock_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(pum_lock_filter_response[n], deg=True),
                np.angle(self.pum_lock_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(pum_drive_align_filter_response[n]),
                                   np.abs(
                                       self.pum_drive_align_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(pum_drive_align_filter_response[n], deg=True),
                np.angle(self.pum_drive_align_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(uim_lock_filter_response[n]),
                                   np.abs(self.uim_lock_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(uim_lock_filter_response[n], deg=True),
                np.angle(self.uim_lock_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(uim_drive_align_filter_response[n]),
                                   np.abs(
                                       self.uim_drive_align_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(uim_drive_align_filter_response[n], deg=True),
                np.angle(self.uim_drive_align_filter_response[n], deg=True))


class TestSUSDigitalFiltersResponse(unittest.TestCase):

    def setUp(self):
        # Pre-defined values
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.TST_digital_filter_response = np.array(
            [-0.03129874344172154+0.30517065377704694j,
             2.4260093192870076+0.5395598117548245j,
             4.746279555953268-14.440418228455206j,
             -10.900389910048899-21.609152361943185j,
             -31.403510154862442-18.172294970461078j,
             -39.725084402162665-5.666968424942829j,
             -36.05682088340858+2.638797989766907j,
             -35.07010453075894+7.434176388352742j,
             -22.901848297541846+31.97868922991029j,
             -25.240115692074934-28.531649366848185j])
        self.PUM_digital_filter_response = np.array(
            [-0.15850099452070285+0.12925881299779612j,
             -2.9964600482668753+0.30212842737343387j,
             -30.56202717042201+13.53346379254984j,
             -117.79715294911026+133.5495986262722j,
             65.71236422447501+691.9714772832681j,
             2010.62556364388+287.8260317781964j,
             -445.7019123170684-210.3550825010501j,
             66.70979808782154-184.31083381342648j,
             -92.35598766208558-56.12007069513851j,
             -3.187299362497145-11.170597438197754j])
        self.UIM_digital_filter_response = np.array(
            [0.0007693760416669129+0.0021754411075238186j,
             -0.015583440206531879-0.014783007553054835j,
             -0.4609157605911597-0.7647563380084396j,
             -17.58562534776213-2.6979597757334934j,
             -15.023972869782675+61.1697093779249j,
             84.18892695637918+136.61857120681844j,
             175.86077619832767-194.04205619047102j,
             22.11561257101802-78.64207839731334j,
             -30.972277953914173-3.640637204710241j,
             -1.168041821378937-2.944313819270134j])

    def tearDown(self):
        del self.frequencies
        del self.TST_digital_filter_response
        del self.PUM_digital_filter_response
        del self.UIM_digital_filter_response

    def test_sus_digital_filters_response(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0
''',
                                            measurement='actuation_x_arm')
        [uim, pum, tst] = A.sus_digital_filters_response(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(uim[n]),
                                   np.abs(self.UIM_digital_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(uim[n], deg=True),
                np.angle(self.UIM_digital_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(pum[n]),
                                   np.abs(self.PUM_digital_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(pum[n], deg=True),
                np.angle(self.PUM_digital_filter_response[n], deg=True))
            self.assertAlmostEqual(np.abs(tst[n]),
                                   np.abs(self.TST_digital_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(tst[n], deg=True),
                np.angle(self.TST_digital_filter_response[n], deg=True))


class TestSUSDigitalCompensationResponse(unittest.TestCase):

    def setUp(self):
        # Pre-defined values
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_digital_response = {
            'UIM': {
                'UL': np.ones(len(self.frequencies), dtype='complex128'),
                'LL': np.ones(len(self.frequencies), dtype='complex128'),
                'UR': np.ones(len(self.frequencies), dtype='complex128'),
                'LR': np.ones(len(self.frequencies), dtype='complex128')},
            'PUM': {
                'UL': np.ones(len(self.frequencies), dtype='complex128'),
                'LL': np.ones(len(self.frequencies), dtype='complex128'),
                'UR': np.ones(len(self.frequencies), dtype='complex128'),
                'LR': np.ones(len(self.frequencies), dtype='complex128')},
            'TST': {
                'UL': np.ones(len(self.frequencies), dtype='complex128'),
                'LL': np.ones(len(self.frequencies), dtype='complex128'),
                'UR': np.ones(len(self.frequencies), dtype='complex128'),
                'LR': np.ones(len(self.frequencies), dtype='complex128')}}

    def tearDown(self):
        del self.frequencies
        del self.known_digital_response

    def test_sus_digital_filters_response(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_filter_bank_UL =
tst_compensation_filter_modules_UL =
tst_filter_gain_UL =
tst_front_end_driver_compensation_UL = ON
''',
                                            measurement='actuation_x_arm')

        test_response = A.sus_digital_compensation_response(self.frequencies)
        for i, stage in enumerate(test_response.keys()):
            for j, quadrant in enumerate(test_response[stage].keys()):
                for k, val in enumerate(test_response[stage][quadrant]):
                    self.assertAlmostEqual(
                        np.abs(val),
                        np.abs(self.known_digital_response[stage][quadrant][k]))
                    self.assertAlmostEqual(
                        np.angle(val, deg=True),
                        np.angle(self.known_digital_response[stage][quadrant][k], deg=True))


class TestCombineSusQuadrants(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_response_uim = np.ones(10, dtype='complex128')
        self.known_response_pum = np.ones(10, dtype='complex128')
        self.known_response_tst = np.array(
            [0.9999999466257983-0.00016911662744827964j,
             0.9999996457380687-0.0004356951588073407j,
             0.9999976486556987-0.0011224778884255887j,
             0.9999843937348445-0.002891770325343889j,
             0.9998964328311806-0.007448857795332509j,
             0.9993133274150832-0.01916976739782008j,
             0.9954745224635373-0.0490354097861226j,
             0.9713081800603385-0.12062585856519095j,
             0.853212124266505-0.2389308400121405j,
             0.6141101867460235-0.23978652937998482j])

    def tearDown(self):
        del self.frequencies
        del self.known_response_uim
        del self.known_response_pum
        del self.known_response_tst

    def test_combine_sus_quadrants(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR =
tst_driver_uncompensated_Z_LR =
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR =
tst_driver_uncompensated_P_LR =
''',
                                            measurement='actuation_x_arm')

        uim, pum, tst = A.combine_sus_quadrants(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(uim[n]),
                np.abs(self.known_response_uim[n]))
            self.assertAlmostEqual(
                np.angle(uim[n], deg=True),
                np.angle(self.known_response_uim[n], deg=True))
            self.assertAlmostEqual(
                np.abs(pum[n]),
                np.abs(self.known_response_pum[n]))
            self.assertAlmostEqual(
                np.angle(pum[n], deg=True),
                np.angle(self.known_response_pum[n], deg=True))
            self.assertAlmostEqual(
                np.abs(tst[n]),
                np.abs(self.known_response_tst[n]))
            self.assertAlmostEqual(
                np.angle(tst[n], deg=True),
                np.angle(self.known_response_tst[n], deg=True))


class TestMatlabForce2LengthResponse(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.TST_known_F2L_response = np.array(
            [-0.00041148596021480487-0.0015143517370263232j,
             -0.00010354107678528522-1.1765319216695903e-07j,
             -1.4644930853742117e-05-3.5466075041766223e-09j,
             -2.188160606001966e-06-2.0395742130086124e-10j,
             -3.292677675426089e-07-1.1898037360168971e-11j,
             -4.9599253544478063e-08-6.955427397606517e-13j,
             -7.47255300169034e-09-4.067318835723566e-14j,
             -1.1258397964013227e-09-2.378592628361941e-15j,
             -1.6961994305038427e-10-1.390972048985424e-16j,
             -2.5555839369830415e-11-8.134681297265418e-18j])
        self.PUM_known_F2L_response = np.array(
            [-0.0009424811357710207+0.0019062104335214956j,
             1.0293251164397253e-05+7.073109949162567e-07j,
             1.4774003935897678e-07+7.750438534956482e-11j,
             3.1997187919682874e-09+6.030048116065241e-13j,
             7.28815782626994e-11+5.275371501679252e-15j,
             1.771401646942218e-12+4.970091549924512e-17j,
             6.80485101377138e-14+7.4184509850763135e-19j,
             -3.939882987588383e-15-1.656518616222811e-20j,
             -2.4980183082476873e-15-3.735923277186558e-21j,
             9.851007172567447e-18+6.41299576873862e-24j])
        self.UIM_known_F2L_response = np.array(
            [-0.0004539112097792617+0.0016706807311863258j,
             -3.270608446356798e-06-2.0483033482849557e-06j,
             -1.0732718463641292e-08+9.358318080529727e-13j,
             -3.040012769594837e-11+1.284586711941435e-14j,
             -1.142690304598126e-13+2.062418770178432e-16j,
             6.853638405500286e-16+1.2429532362755387e-16j,
             3.1125363563944354e-17-8.449710365518706e-19j,
             9.159264149563957e-22+1.3178814099168241e-23j,
             -2.6561794705370495e-25-7.338344152095596e-27j,
             -1.1726221364201392e-30-2.4464769898562298e-30j])

    def tearDown(self):
        del self.frequencies
        del self.TST_known_F2L_response
        del self.PUM_known_F2L_response
        del self.UIM_known_F2L_response

    def test_matlab_force2length_response(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
''',
                                            measurement='actuation_x_arm')
        [uim, pum, tst] = A.matlab_force2length_response(
            'test/H1susdata_O3.mat', self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(uim[n]),
                                   np.abs(self.UIM_known_F2L_response[n]))
            self.assertAlmostEqual(
                np.angle(uim[n], deg=True),
                np.angle(self.UIM_known_F2L_response[n], deg=True))
            self.assertAlmostEqual(np.abs(pum[n]),
                                   np.abs(self.PUM_known_F2L_response[n]))
            self.assertAlmostEqual(
                np.angle(pum[n], deg=True),
                np.angle(self.PUM_known_F2L_response[n], deg=True))
            self.assertAlmostEqual(np.abs(tst[n]),
                                   np.abs(self.TST_known_F2L_response[n]))
            self.assertAlmostEqual(
                np.angle(tst[n], deg=True),
                np.angle(self.TST_known_F2L_response[n], deg=True))


class TestDigitalOutToDisplacement(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.TST_known_tf = np.array(
            [1.9381128366288105e-15+7.084923321696425e-15j,
             4.846416041981388e-16-1.5765449186707969e-18j,
             6.854411598086494e-17-7.585382386850981e-19j,
             1.0237678592541069e-17-2.9738538652980284e-19j,
             1.5367561509460086e-18-1.1550321717363778e-19j,
             2.2772458041669347e-19-4.4586564579385514e-20j,
             3.063100975597829e-20-1.665633663061682e-20j,
             1.472512511979171e-21-4.948808065557348e-21j,
             -6.970960601188869e-22+1.0099474651594038e-22j,
             -3.8963114263552186e-23-5.016546730668483e-23j])
        self.PUM_known_tf = np.array(
            [-5.618021460861343e-13+1.1401945563127125e-12j,
             6.154175662825378e-15+4.010967874186301e-16j,
             8.830745619372059e-17-7.557564507542089e-19j,
             1.9121021132528307e-18-4.4390085030266175e-20j,
             4.3486514554534227e-20-2.6215994978040016e-21j,
             1.0462587696100501e-21-1.6379277822286195e-22j,
             3.749784155222869e-23-1.5864846571311978e-23j,
             -1.2139274103725887e-24+2.0358371914030717e-24j,
             1.3971705879842376e-24+6.693133108398555e-25j,
             2.6080357979757855e-27-5.232778377591158e-27j])
        self.UIM_known_tf = np.array(
            [-3.430763161527003e-11+1.2696021299733796e-10j,
             -2.4899857500856507e-13-1.547223638897765e-13j,
             -8.152763544441236e-16+7.476151512063072e-18j,
             -2.3086963982589343e-18+5.500858196896584e-20j,
             -8.663921746539537e-21+5.386329216110624e-22j,
             5.290544966557914e-23+1.2746440643491213e-24j,
             2.154672382803322e-24-9.81407533663976e-25j,
             3.672997786320173e-29-5.963176342634914e-29j,
             1.863046548604978e-32+9.566192650734058e-33j,
             -2.0460864930999043e-37-3.153188911481507e-39j])

    def tearDown(self):
        del self.frequencies
        del self.TST_known_tf
        del self.PUM_known_tf
        del self.UIM_known_tf

    def test_digital_out_to_displacement(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
''',
                                            measurement='actuation_x_arm')
        [uim_response, pum_response, tst_response] = \
            A.digital_out_to_displacement(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(uim_response[n]),
                                   np.abs(self.UIM_known_tf[n]))
            self.assertAlmostEqual(
                np.angle(uim_response[n], deg=True),
                np.angle(self.UIM_known_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(pum_response[n]),
                                   np.abs(self.PUM_known_tf[n]))
            self.assertAlmostEqual(
                np.angle(pum_response[n], deg=True),
                np.angle(self.PUM_known_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(tst_response[n]),
                                   np.abs(self.TST_known_tf[n]))
            self.assertAlmostEqual(
                np.angle(tst_response[n], deg=True),
                np.angle(self.TST_known_tf[n], deg=True))


class TestDrivealignToDisplacement(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.model_string = '''
[metadata]
[interferometer]
[actuation_x_arm]
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0
'''
        self.sign_string = "darm_feedback_sign = -1"
        self.TST_known_long_tf = np.array(
            [-2.222881500981626e-15+3.691011841669097e-16j,
             1.1764193182970055e-15+2.5849635956379406e-16j,
             3.1617534435707367e-16-9.92839053775603e-16j,
             -1.1700305084971196e-16-2.1853458973419276e-16j,
             -5.006230661758991e-17-2.4902470964105e-17j,
             -9.308653195938007e-18+1.9222273202986028e-19j,
             -1.1111368291657706e-18+5.936117452101705e-19j,
             -5.412995749774128e-20+1.7611360028653088e-19j,
             2.4897361921808887e-20-3.511338429355629e-21j,
             1.3890965613664794e-21+1.7923601565333542e-21j])
        self.PUM_known_long_tf = np.array(
            [-5.636311671296026e-13+1.1503626120207196e-12j,
             6.539680039025585e-15+3.344573000972991e-16j,
             9.436017253315024e-17-3.810611139309467e-17j,
             1.2201453005199638e-18-1.1859853195589023e-18j,
             2.3160192732375865e-21-2.230152515596516e-20j,
             -1.7088823698982324e-22-1.4723824048886304e-22j,
             -3.9895135060052445e-25+2.9092246327605506e-25j,
             1.1353220905153887e-26+8.382906433078052e-27j,
             3.841572759246625e-27-3.1425560227349834e-27j,
             -4.588083981792824e-30-2.906408489212469e-30j])
        self.UIM_known_long_tf = np.array(
            [-3.430763161527003e-11+1.2696021299733796e-10j,
             -2.4899857500856507e-13-1.547223638897765e-13j,
             -8.152763544441236e-16+7.476151512063072e-18j,
             -2.3086963982589343e-18+5.500858196896584e-20j,
             -8.663921746539537e-21+5.386329216110624e-22j,
             5.290544966557914e-23+1.2746440643491213e-24j,
             2.154672382803322e-24-9.81407533663976e-25j,
             3.672997786320173e-29-5.963176342634914e-29j,
             1.863046548604978e-32+9.566192650734058e-33j,
             -2.0460864930999043e-37-3.153188911481507e-39j])
        signval = float(self.sign_string.split(' = ')[-1])
        self.TST_known_darm_tf = signval * self.TST_known_long_tf
        self.PUM_known_darm_tf = signval * self.PUM_known_long_tf
        self.UIM_known_darm_tf = signval * self.UIM_known_long_tf

    def tearDown(self):
        del self.frequencies
        del self.model_string
        del self.sign_string
        del self.TST_known_long_tf
        del self.PUM_known_long_tf
        del self.UIM_known_long_tf
        del self.TST_known_darm_tf
        del self.PUM_known_darm_tf
        del self.UIM_known_darm_tf

    def test_drivealign_to_longitudinal_displacement(self):
        A = pydarm.actuation.ActuationModel(self.model_string,
                                            measurement='actuation_x_arm')
        [uim_response, pum_response, tst_response] = \
            A.drivealign_to_longitudinal_displacement(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(uim_response[n]),
                                   np.abs(self.UIM_known_long_tf[n]))
            self.assertAlmostEqual(
                np.angle(uim_response[n], deg=True),
                np.angle(self.UIM_known_long_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(pum_response[n]),
                                   np.abs(self.PUM_known_long_tf[n]))
            self.assertAlmostEqual(
                np.angle(pum_response[n], deg=True),
                np.angle(self.PUM_known_long_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(tst_response[n]),
                                   np.abs(self.TST_known_long_tf[n]))
            self.assertAlmostEqual(
                np.angle(tst_response[n], deg=True),
                np.angle(self.TST_known_long_tf[n], deg=True))

    def test_drivealign_to_darm_displacement(self):
        A = pydarm.actuation.ActuationModel(
            self.model_string + self.sign_string,
            measurement='actuation_x_arm')
        [uim_response, pum_response, tst_response] = \
            A.drivealign_to_darm_displacement(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(uim_response[n]),
                                   np.abs(self.UIM_known_darm_tf[n]))
            self.assertAlmostEqual(
                np.angle(uim_response[n], deg=True),
                np.angle(self.UIM_known_darm_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(pum_response[n]),
                                   np.abs(self.PUM_known_darm_tf[n]))
            self.assertAlmostEqual(
                np.angle(pum_response[n], deg=True),
                np.angle(self.PUM_known_darm_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(tst_response[n]),
                                   np.abs(self.TST_known_darm_tf[n]))
            self.assertAlmostEqual(
                np.angle(tst_response[n], deg=True),
                np.angle(self.TST_known_darm_tf[n], deg=True))


class TestDrivealignToDisplacementForMeasurement(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.TST_known_tf = np.array(
            [5.021191554058338e-05-8.337501336501237e-06j,
             -2.65737365777503e-05-5.83908650471638e-06j,
             -7.1419775097599665e-06+2.2426904309365334e-05j,
             2.642942192222995e-06+4.9364036533587704e-06j,
             1.1308404476528101e-06+5.625134620308336e-07j,
             2.1027000668484317e-07-4.3420540327504015e-09j,
             2.5099092594663896e-08-1.3408894176873065e-08j,
             1.222723232386295e-09-3.9781703249724626e-09j,
             -5.623980556089653e-10+7.931643165474655e-11j,
             -3.137783061591325e-11-4.0487015056095644e-11j])
        self.PUM_known_tf = np.array(
            [1.912559101220233e-11-3.903503943063181e-11j,
             -2.2190974004158754e-13-1.1349077030787211e-14j,
             -3.20190609206482e-15+1.293047553209863e-15j,
             -4.140296235222137e-17+4.024381810515447e-17j,
             -7.858904897311117e-20+7.567534834056722e-19j,
             5.798718594836214e-21+4.996207685404243e-21j,
             1.3537541588073445e-23-9.871817552631662e-24j,
             -3.852467222651472e-25-2.8445559664330002e-25j,
             -1.3035537018142603e-25+1.0663576595639575e-25j,
             1.5568659592103235e-28+9.862261585383335e-29j])
        self.UIM_known_tf = np.array(
            [2.0996102579724625e-11-7.769902876214075e-11j,
             1.523859088179713e-13+9.468932918590974e-14j,
             4.989451373587049e-16-4.575368122437622e-18j,
             1.4129108924473284e-18-3.366498284514433e-20j,
             5.302277690660671e-21-3.29640710900283e-22j,
             -3.237787617232506e-23-7.800759267742482e-25j,
             -1.3186489490840405e-24+6.006166056695082e-25j,
             -2.2478566623746473e-29+3.6494347262147576e-29j,
             -1.1401753663433159e-32-5.8544630665447124e-33j,
             1.2521949162178118e-37+1.929736175937275e-39j])

    def tearDown(self):
        del self.frequencies
        del self.TST_known_tf
        del self.PUM_known_tf
        del self.UIM_known_tf

    def test_known_actuation_terms_for_measurement(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0
''',
                                            measurement='actuation_x_arm')
        [uim_response, pum_response, tst_response] = \
            A.known_actuation_terms_for_measurement(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(uim_response[n]),
                                   np.abs(self.UIM_known_tf[n]))
            self.assertAlmostEqual(
                np.angle(uim_response[n], deg=True),
                np.angle(self.UIM_known_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(pum_response[n]),
                                   np.abs(self.PUM_known_tf[n]))
            self.assertAlmostEqual(
                np.angle(pum_response[n], deg=True),
                np.angle(self.PUM_known_tf[n], deg=True))
            self.assertAlmostEqual(np.abs(tst_response[n]),
                                   np.abs(self.TST_known_tf[n]))
            self.assertAlmostEqual(
                np.angle(tst_response[n], deg=True),
                np.angle(self.TST_known_tf[n], deg=True))


class TestComputeActuationSingleStage(unittest.TestCase):

    def setUp(self):
        # TST stage
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_single_stage = np.array(
            [-2.2227711784771e-15+3.6970596409765503e-16j,
             1.176595688578433e-15+2.5766842006473456e-16j,
             3.1437592695198093e-16-9.934059363981118e-16j,
             -1.1802093456019643e-16-2.1798593987209916e-16j,
             -5.035849592429691e-17-2.4299179619229708e-17j,
             -9.299048831148297e-18+4.806970343138081e-19j,
             -1.0605041246314987e-18+6.814035934325136e-19j,
             -1.4850855646504008e-20+1.84502133909869e-19j,
             1.2735108604970117e-20-2.4605184633493507e-20j,
             -4.47870011782826e-22+2.3778641129758554e-21j])

    def tearDown(self):
        del self.frequencies
        del self.known_single_stage

    def test_compute_actuation_single_stage(self):
        """ Test a single actuation stage """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0
''',
                                            measurement='actuation_x_arm')
        single_stage = A.compute_actuation_single_stage(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(single_stage[n]),
                                   np.abs(self.known_single_stage[n]))
            self.assertAlmostEqual(
                np.angle(single_stage[n], deg=True),
                np.angle(self.known_single_stage[n], deg=True))


class TestActuationStageResidual(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_residual = np.array(
            [0.9902575615036632-0.0016871227384657537j,
             0.9902492531426801-0.004346523822516416j,
             0.9901941084811342-0.011197736013719756j,
             0.9898281244071855-0.028844992443191404j,
             0.9874002528379973-0.07424928569758847j,
             0.9713421858209306-0.19019480399674393j,
             0.8672208231254349-0.47157796687119996j,
             0.27670562456405506-0.9299571986485128j,
             -0.8694694693325025+0.1259688027976555j,
             -0.32255394920871006-0.4152917257163378j])

    def tearDown(self):
        del self.frequencies
        del self.known_residual

    def test_actuation_stage_residual(self):
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0
''',
                                            measurement='actuation_x_arm')
        residual = A.actuation_stage_residual(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(residual[n]),
                                   np.abs(self.known_residual[n]))
            self.assertAlmostEqual(
                np.angle(residual[n], deg=True),
                np.angle(self.known_residual[n], deg=True))


class TestDARMOutputMatrixValues(unittest.TestCase):

    def setUp(self):
        # darm_output_matrix = 1.0, -1.0, 0.0, 0.0
        # darm_feedback_x    = OFF, ON, ON, ON
        # darm_feedback_y    = OFF, OFF, OFF, OFF
        self.known_output_matrix = np.array([[0, 1.0, 1.0, 1.0],
                                             [0, 0, 0, 0]])

    def tearDown(self):
        del self.known_output_matrix

    def test_output_matrix_values(self):
        A = pydarm.actuation.DARMActuationModel('''
[metadata]
[interferometer]
[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF
''')
        output_matrix = A.darm_output_matrix_values()
        for n in range(len(output_matrix[:, 0])):
            for m in range(len(output_matrix[0, :])):
                self.assertAlmostEqual(output_matrix[n, m],
                                       self.known_output_matrix[n, m])


class TestComputeActuation(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.model_string = '''
[metadata]
[interferometer]
[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF
'''
        self.known_A = np.array(
            [-3.63234851552863e-13-2.297849477733715e-13j,
             -1.5785405207037768e-14+7.022803250131525e-15j,
             -1.9906028977349535e-15+8.499087530441956e-16j,
             -2.9626430309517743e-16+4.980882746237151e-17j,
             -4.5497245018124926e-17+5.851368523167505e-18j,
             -7.117215795049456e-18+7.70258488252587e-19j,
             -9.974890659056889e-19+7.966180520139261e-19j,
             3.8712043816670766e-20+1.8134804487474165e-19j,
             -7.463095731015656e-21-2.67674977773168e-20j,
             2.3885824821188465e-21-3.8665176935995385e-22j])

    def tearDown(self):
        del self.frequencies
        del self.model_string
        del self.known_A

    def test_compute_actuation(self):
        A = pydarm.actuation.DARMActuationModel(self.model_string)
        actuation_tf = A.compute_actuation(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(actuation_tf[n]),
                                   np.abs(self.known_A[n]))
            self.assertAlmostEqual(
                np.angle(actuation_tf[n], deg=True),
                np.angle(self.known_A[n], deg=True))

    def test_arm_super_actuator(self):
        A = pydarm.actuation.DARMActuationModel(self.model_string)
        actuation_tf = A.arm_super_actuator(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(actuation_tf[n]),
                                   np.abs(self.known_A[n]))
            self.assertAlmostEqual(
                np.angle(actuation_tf[n], deg=True),
                np.angle(self.known_A[n], deg=True))


class TestComputeMultiArmActuation(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.model_string = '''
[metadata]
[interferometer]
[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation_y_arm]
darm_feedback_sign = 1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_front_end_driver_compensation = ON
pum_front_end_driver_compensation = ON
uim_front_end_driver_compensation = ON
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, ON, ON, ON
'''
        self.known_A = 2*np.array(
            [-3.63234851552863e-13-2.297849477733715e-13j,
             -1.5785405207037768e-14+7.022803250131525e-15j,
             -1.9906028977349535e-15+8.499087530441956e-16j,
             -2.9626430309517743e-16+4.980882746237151e-17j,
             -4.5497245018124926e-17+5.851368523167505e-18j,
             -7.117215795049456e-18+7.70258488252587e-19j,
             -9.974890659056889e-19+7.966180520139261e-19j,
             3.8712043816670766e-20+1.8134804487474165e-19j,
             -7.463095731015656e-21-2.67674977773168e-20j,
             2.3885824821188465e-21-3.8665176935995385e-22j])
        self.known_super_act = np.array(
            [-4.4452584691383385e-15+7.411167179232215e-16j,
             2.3536993809387682e-15+5.130116377710275e-16j,
             6.236926239205679e-16-1.9884058495535515e-15j,
             -2.388957355496579e-16-4.3441463078690887e-16j,
             -1.0152362687187823e-16-4.6889936601361595e-17j,
             -1.8538651825179792e-17+1.769718541540502e-18j,
             -1.955189357592536e-18+1.5915885140228738e-18j,
             7.665512735370611e-20+3.621744507461856e-19j,
             -1.4601630153846298e-20-5.34526407669915e-20j,
             4.777143021044974e-21-7.734375898626229e-22j])
        self.known_super_act_drivealign = 2*np.array(
            [2.222881500981626e-15-3.691011841669097e-16j,
             -1.1764193182970055e-15-2.5849635956379406e-16j,
             -3.1617534435707367e-16+9.92839053775603e-16j,
             1.1700305084971196e-16+2.1853458973419276e-16j,
             5.006230661758991e-17+2.4902470964105e-17j,
             9.308653195938007e-18-1.9222273202986028e-19j,
             1.1111368291657706e-18-5.936117452101705e-19j,
             5.412995749774128e-20-1.7611360028653088e-19j,
             -2.4897361921808887e-20+3.511338429355629e-21j,
             -1.3890965613664794e-21-1.7923601565333542e-21j])

    def tearDown(self):
        del self.frequencies
        del self.model_string
        del self.known_A
        del self.known_super_act
        del self.known_super_act_drivealign

    def test_multi_actuation(self):
        A = pydarm.actuation.DARMActuationModel(self.model_string)
        actuation_tf = A.compute_actuation(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(actuation_tf[n]),
                                   np.abs(self.known_A[n]))
            self.assertAlmostEqual(
                np.angle(actuation_tf[n], deg=True),
                np.angle(self.known_A[n], deg=True))

    def test_stage_super_actuator(self):
        A = pydarm.actuation.DARMActuationModel(self.model_string)
        actuation_tf = A.stage_super_actuator(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(actuation_tf[n]),
                                   np.abs(self.known_super_act[n]))
            self.assertAlmostEqual(
                np.angle(actuation_tf[n], deg=True),
                np.angle(self.known_super_act[n], deg=True))

    def test_stage_super_actuator_drivealign(self):
        A = pydarm.actuation.DARMActuationModel(self.model_string)
        actuation_tf = A.stage_super_actuator_drivealign(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(np.abs(actuation_tf[n]),
                                   np.abs(self.known_super_act_drivealign[n]))
            self.assertAlmostEqual(
                np.angle(actuation_tf[n], deg=True),
                np.angle(self.known_super_act_drivealign[n], deg=True))


if __name__ == '__main__':
    unittest.main()
